"""
The basic sprite ("DesignerObject") and subclass sprites (image, circle, rectangle, text, various widgest, etc.).
"""

from designer.objects.designer_object import DesignerObject
from designer.objects.arc import arc
from designer.objects.circle import circle
from designer.objects.ellipse import ellipse
from designer.objects.group import group
from designer.objects.image import image
from designer.objects.line import line
from designer.objects.rectangle import rectangle
from designer.objects.shape import shape, lines
from designer.objects.text import text
from designer.objects.emoji import emoji, get_emoji_name, set_emoji_name
from designer.objects.pen import pen


from typing import Optional, Union

import json
from urllib.request import urlopen, Request
import pygame
import math
import sys
import os
import io

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities import Vec2D
from designer.utilities.util import _anchor_offset
from designer.utilities.gif_image import GifImage


try:
    from zipfile import ZipFile
    from os import path
    _THIS_DIRECTORY = path.abspath(path.dirname(__file__))
    _UNICODE_LIST_PATH = path.join(_THIS_DIRECTORY, '../data/unicode_names.json')
    _EMOJI_DATABASE = path.join(_THIS_DIRECTORY, '../data/emojis.zip')
    ALT_MODE = False
except:
    # TODO: Emoji support limited in Skuplt
    ALT_MODE = True
    _UNICODE_LIST_PATH = 'src/lib/designer/data/unicode_names.json'

_UNICODE_LOOKUP = {}


def lookup_unicode(name):
    if not _UNICODE_LOOKUP:
        with open(_UNICODE_LIST_PATH) as unicode_list_file:
            _UNICODE_LOOKUP.update(json.load(unicode_list_file))
    return _UNICODE_LOOKUP[name.lower()].lower()


class Emoji(DesignerObject):
    FIELDS = (*DesignerObject.FIELDS, 'name')
    DEFAULT_EMOJI_SIZE = 36
    _EMOJI_CACHE = {}

    def __init__(self, center, name, anchor, **kwargs):
        """
        Creates Image Designer Object on window

        :param center: x, y coordinates of center of circle
        :type center: Tuple[int]
        :param anchor: the anchor to draw the circle at
        :type anchor: str
        :param name: The name of the emoji
        :type name: str
        :param width: width of image in pixels
        :type width: int
        :param height: height of image in pixels
        :type height: int
        """
        super().__init__(**kwargs)

        x, y = center
        x = x if x is not None else get_width() / 2
        y = y if y is not None else get_height() / 2
        center = x, y

        self._pos = center
        self._anchor = anchor
        # Image specific data
        self._size = Vec2D(self.DEFAULT_EMOJI_SIZE, self.DEFAULT_EMOJI_SIZE)
        self._name = name
        # Internal field holding the raw svg information
        self._svg: Optional[Union[str, dict]] = None
        #: Internal field holding the original version of the image
        self._internal_image: Optional[InternalImage] = None
        self._internal_image_version: Optional[int] = None

        self._load_image()

        # And draw!
        self._redraw_internal_image()
        self._recalculate_offset()

    def _get_unicode(self, text):
        # Look up the name or character
        try:
            return lookup_unicode(text)
        except KeyError:
            character = text
        try:
            return hex(ord(character))[2:]
        except TypeError:
            raise Exception(f"Unknown unicode name '{text}', could not find an emoji for this name.")

    def _load_image(self):
        if self._name in self._EMOJI_CACHE:
            self._svg = self._EMOJI_CACHE[self._name]
        target = self._get_unicode(self._name)
        try:
            emoji_database = ZipFile(_EMOJI_DATABASE)
            with emoji_database.open(target) as svg_file:
                self._svg = svg_file.read()
            self._EMOJI_CACHE[self._name] = self._svg
        except KeyError:
            raise Exception(f"Unknown unicode name '{self._name}', could not find an emoji for this name.")

    def _alt_load_image(self):
        target = self._get_unicode(self._name)
        # Hackish skulpt-specific solution for loading svgs in a special way
        self._svg = {'code': target}

    def _recalculate_offset(self):
        """
        Recalculates this designer object's offset based on its position, transform
        offset, anchor, its image, and the image's scaling.
        """
        if self._internal_image is None:
            return
        size = self._internal_image.size
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._size = size
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        """
        Calculates the transforms that need to be applied to this designer object's
        image. In order: flipping, scaling, and rotation.
        """
        # Transform the SVG
        l, r = Vec2D(self.DEFAULT_EMOJI_SIZE, self.DEFAULT_EMOJI_SIZE) * self._scale
        flip_x, flip_y = "-" if self._flip_x else "", "-" if self._flip_y else ""
        transforms = f"rotate({self._angle})\nscale({flip_x}{self._scale[0]}, {flip_y}{self._scale[1]})"
        image_data = (f'<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 {l} {r}">'
                      f'<g transform="{transforms}">{self._svg}'
                      f'</g></svg>'.encode())
        image_file = io.BytesIO(image_data)
        # Renders it into the image
        self._internal_image = InternalImage(filename=self._name+".svg", fileobj=image_file)
        source = self._internal_image._surf
        # Finish updates
        self._transform_image = source
        self._recalculate_offset()
        self._expire_static()

    def _alt_redraw_internal_image(self):
        """
        Calculates the transforms that need to be applied to this designer object's
        image. In order: flipping, scaling, and rotation.

        This is the alternate version used by Skulpt.
        """
        self._svg['angle'] = self._angle
        self._svg['flip_x'] = "-" if self.flip_x else ""
        self._svg['flip_y'] = "-" if self.flip_y else ""
        self._svg['scale_x'] = self._scale[0]
        self._svg['scale_y'] = self._scale[1]
        self._svg['width'] = self.DEFAULT_EMOJI_SIZE #self._size[0]
        self._svg['height'] = self.DEFAULT_EMOJI_SIZE #self._size[1]
        # Renders it into the image
        self._internal_image = InternalImage(self._svg)
        source = self._internal_image._surf
        # Flip
        if self._flip_x or self._flip_y:
            source = pygame.transform.flip(source, self._flip_x, self._flip_y)
        # Scale
        #if self._scale != (1.0, 1.0):
        #    new_size = self._scale * self._internal_image.size
        #    new_size = (int(new_size[0]), int(new_size[1]))
        #    if 0 in new_size:
        #        return self._make_blank_surface()
        #    new_surf = DesignerSurface(new_size)
        #    source = pygame.transform.smoothscale(source, new_size, new_surf)
        # Rotate

        if self._angle != 0:
            angle = self._angle % 360
            old = Vec2D(source.get_rect().center)
            source = pygame.transform.rotate(source, angle).convert_alpha()
            new = source.get_rect().center
            self._transform_offset = old - new
        else:
            self._transform_offset = Vec2D(0, 0)

        # Finish updates
        self._transform_image = source
        self._recalculate_offset()
        self._expire_static()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value == self._name:
            return
        self._name = value
        self._load_image()
        self._redraw_internal_image()
        self._recalculate_offset()
        self._expire_static()


def emoji(name, x=None, y=None, anchor='center', **kwargs):
    """
    Function to create an image from the emoji with the given name.

    :param name: The emoji's name
    :type name: str
    :param args: left top corner of image and width and height of image
    :type args: two Tuples (left, top), (width, height) or four ints left, top, width, height
    :return: Image object
    """
    if x is not None and y is None:
        if isinstance(x, (list, tuple)):
            x, y = x
    return Emoji((x, y), name, anchor, **kwargs)


if ALT_MODE:
    Emoji._load_image = Emoji._alt_load_image
    Emoji._redraw_internal_image = Emoji._alt_redraw_internal_image


def get_emoji_name(an_emoji: Emoji) -> str:
    return an_emoji.name


def set_emoji_name(an_emoji: Emoji, new_name: str):
    an_emoji.name = new_name




"""When they have no other home, functions and classes are added here.
Eventually, they should be refactored to a more permanent home."""
import difflib

import designer
import math
import pygame

from designer.utilities.argument_checks import make_suggestions
from designer.utilities.memoize import _ImageMemoize


ANCHOR_WORDS = ('topleft', 'topright', 'midtop', 'bottomleft', 'bottomright', 'midbottom', 'midleft', 'midright', 'center')


def _anchor_offset(anchor, width, height):
    """
    Given an `anchor` position (either a string or a 2-tuple position), finds
    the correct offset in a rectangle of size (`width`, `height`). If the
    `anchor` is a 2-tuple (or Vec2D), then it multiplies both components by -1.

    >>> anchor_offset("topleft", 100, 100)
    Vec2D(0,0)
    >>> anchor_offset("bottomright", 100, 100)
    Vec2D(100,100)
    >>> anchor_offset("center", 100, 100)
    Vec2D(50,50)
    >>> anchor_offset((10, 10), 100, 100)
    Vec2D(-10,-10)

    For a complete list of the anchor positions, see `Anchor Offset Lists`_.

    :param anchor: The (possibly named) position to offset by.
    :type anchor: string or :class:`Vec2D <spyral.Vec2D>`
    :param width: the width of the rectangle to offset in.
    :type width: int
    :param height: the height of the rectangle to offset in.
    :type height: int
    :rtype: :class:`Vec2D <spyral.Vec2D>`
    """
    w = width
    h = height
    a = anchor
    if a == 'topleft':
        offset = (0, 0)
    elif a == 'topright':
        offset = (w, 0)
    elif a == 'midtop':
        offset = (w / 2., 0)
    elif a == 'bottomleft':
        offset = (0, h)
    elif a == 'bottomright':
        offset = (w, h)
    elif a == 'midbottom':
        offset = (w / 2., h)
    elif a == 'midleft':
        offset = (0, h / 2.)
    elif a == 'midright':
        offset = (w, h / 2.)
    elif a == 'center':
        offset = (w / 2., h / 2.)
    elif isinstance(a, str):
        suggestions = make_suggestions(a, ANCHOR_WORDS, cutoff=.3)
        if suggestions:
            raise ValueError(f"Invalid anchor {a!r}, perhaps you meant {suggestions}")
        else:
            raise ValueError(f"Invalid anchor {a!r}, should be a string like 'topleft', 'midright', or 'center'.")
    else:
        offset = a * designer.utilities.vector.Vec2D(-1, -1)
    return designer.utilities.vector.Vec2D(offset)


@_ImageMemoize
def scale_surface(s, target_size):
    """
    Internal method to scale a surface `s` by a float `factor`. Uses memoization
    to improve performance.

    :param target_size: The end size of the surface
    :type target_size: :class:`Vec2D <spyral.Vec2D>`
    :returns: A new surface, or the original (both pygame surfaces).
    """
    new_size = (int(math.ceil(target_size[0])),
                int(math.ceil(target_size[1])))
    if new_size == s.get_size():
        return s
    t = pygame.transform.smoothscale(s, new_size,
                                     designer.utilities.surfaces.DesignerSurface(new_size))
    return t


class _Blit:
    """
    An internal class to represent a drawable `surface` with additional data
    (e.g. `rect` representing its location on screen, whether it's `static`).

    .. attribute::surface

        The internal Pygame source surface use to render this blit.

    .. attribute::position

        The current position of this Blit (:class:`Vec2D <spyral.Vec2D>`).

    .. attribute::area

        The portion of the surface to be drawn to the screen
        (:class:`Rect <spyral.Rect>`).

    .. attribute::layer

        The layer of this blit within the scene (`str).

    .. attribute::flags

        Any drawing flags (presently unused).

    .. attribute::static

        Whether this Blit is static.

    .. attribute::final_size

        The final size, set seperately to defer scaling.

    .. attribute::rect

        A rect representing the final position and size of this blit
        (:class:`Rect <spyral.Rect>`).

    """
    __slots__ = ['surface', 'position', 'rect', 'area', 'layer',
                 'flags', 'static', 'clipping', 'final_size', 'id']

    def __init__(self, surface, position, area, layer, flags, static, id):
        self.surface = surface  # pygame surface
        self.position = position  # coordinates to draw at
        self.area = area  # portion of the surface to be drawn to screen
        self.layer = layer  # layer in scene
        self.flags = flags  # any drawing flags (currently unusued)
        self.static = static  # static blits haven't changed
        self.id = id

        # Final size of the surface, the scaling will happen late
        self.final_size = designer.utilities.vector.Vec2D(surface.get_size())
        # Rect is only for finalized sprites
        self.rect = None

    def apply_scale(self, scale):
        """
        Applies the scaling factor to this blit.

        :param scale: The scaling factor
        :type scale: :class:`Vec2D <spyral.Vec2D>`
        """
        self.position = self.position * scale
        self.final_size = self.final_size * scale
        self.area = designer.utilities.rect.Rect(self.area.topleft * scale,
                                                 self.area.size * scale)

    def clip(self, rect):
        """
        Applies any necessary cropping to this blit

        :param rect: The new maximal size of the blit.
        :type rect: :class:`Rect <spyral.Rect>`
        """
        self.area = self.area.clip(designer.utilities.rect.Rect(rect))

    def finalize(self):
        """
        Performs all the final calculations for this blit and calculates the
        rect.
        """
        self.surface = scale_surface(self.surface, self.final_size)
        area = self.area._to_pygame()
        self.surface = self.surface.subsurface(area)
        position = tuple(self.position)
        self.rect = pygame.Rect(position, self.surface.get_size())


class _CollisionBox(object):
    """
    An internal class for managing the collidable area for a sprite or view.
    In many ways, this is a reduced form of a _Blit.

    .. attribute::position

        The current position of this CollisionBox
        (:class:`Vec2D <spyral.Vec2D>`).

    .. attribute::area

        The current offset and size of this CollisionBox.
        (:class:`Rect <spyral.Rect>`).

    .. attribute::rect

        A rect representing the final position and size of this CollisionBox.
        (:class:`Rect <spyral.Rect>`).

    """
    __slots__ = ['position', 'rect', 'area']

    def __init__(self, position, area):
        self.position = position
        self.area = area
        self.rect = None

    def apply_scale(self, scale):
        self.position = self.position * scale
        self.area = designer.utilities.rect.Rect(self.area.topleft * scale,
                                                 self.area.size * scale)

    def clip(self, rect):
        self.area = self.area.clip(designer.utilities.rect.Rect(rect))

    def finalize(self):
        self.rect = designer.utilities.rect.Rect(self.position, self.area.size)


def rect_from_points(points):
    """
    Must be an iterable sequence of pairs (tuple, list, Vec2D)
    :param points:
    :return:
    """
    if not points:
        return designer.utilities.rect.Rect(0, 0, 1, 1)
    left = min(point[0] for point in points)
    top = min(point[1] for point in points)
    return designer.utilities.rect.Rect(
        left, top,
        max(point[0] for point in points) - left,
        max(point[1] for point in points) - top
    )



from typing import List

import pygame

from designer.colors import _process_color
from designer.core.window import Window
from designer.core.event import handle, Event, _pygame_to_spyral, GameEndException, register
from designer.keyboard import KeyboardModule
from designer.mouse import MouseModule
from designer.music import MusicModule
from designer.sfx import SfxModule

DEFAULT_WINDOW_TITLE = "Designer Game"


class Director:
    def __init__(self, width=800, height=600, background_color=(255, 255, 255, 255), fps=30):
        """
        Initializes the Director that will control the game state.

        :param width: width of the game window in pixels
        :type width: int
        :param height: height of the game window in pixels
        :type height: int
        :param background_color: color to initially fill the window with
        """
        pygame.init()
        self._window_size = width, height
        self._window_title = DEFAULT_WINDOW_TITLE
        self._window_color = background_color
        self._fps = fps
        self._tick = 0
        self.running = False
        self.paused = False

        self._windows: List[Window] = []

        self._all_sprites = set()
        self._game_state = None

        self.screen = pygame.display.set_mode(self.window_size)
        self.screen.fill(self._window_color)
        pygame.display.set_caption(self._window_title)

        self.keyboard = KeyboardModule()
        self.mouse = MouseModule()
        self.music = MusicModule()
        self.sfx = SfxModule()

    def _setup_initial_window(self):
        new_window = Window(self._window_size, self._fps)
        self._windows.append(new_window)
        register("system.quit", self.stop)
        new_window._register_default_events()

    def _switch_window(self):
        """
        Ensure that dead sprites are removed from the list and that sprites are
        redrawn on a window change.
        """
        self._all_sprites = {s for s in self._all_sprites
                             if s is not None and s._expire_static()}

    def _track_object(self, object):
        self._all_sprites.add(object)

    def _untrack_object(self, object):
        self._all_sprites.remove(object)

    @property
    def tick(self):
        return self._tick

    @property
    def fps(self):
        return self._fps

    @property
    def current_window(self):
        return self._windows[-1]

    @property
    def window_size(self):
        return self._window_size

    @window_size.setter
    def window_size(self, value):
        self._window_size = value

    @property
    def width(self):
        return self._window_size[0]

    @property
    def height(self):
        return self._window_size[1]

    @property
    def window_title(self):
        return self._window_title

    @window_title.setter
    def window_title(self, value):
        self._window_title = value
        if isinstance(value, str):
            pygame.display.set_caption(value)

    @property
    def window_color(self):
        return self._window_color

    @window_color.setter
    def window_color(self, value):
        self._window_color = value
        self.screen.fill(_process_color(value))
        # TODO: Director shouldn't do this, the window itself should handle that
        self.current_window.background.fill(_process_color(value))
        self.current_window.background._version += 1

    def replace(self, window):
        """
        Replace the currently running window on the stack with *window*.
        Execution will continue after this is called, so make sure you return;
        otherwise you may find unexpected behavior::

            designer.director.replace(Window())
            print("This will be printed!")
            return

        :param window: The new window.
        :type window: :class:`Window <designer.core.window.Window>`
        """
        if self._windows:
            handle('director.scene.exit')
            self._windows.pop()
            self._switch_window()
        self._windows.append(window)
        window._register_default_events()
        handle('director.scene.enter', event=Event(window=window))
        # Empty all events!
        pygame.event.get()

    def pop(self):
        """
        Pop the top window off the stack, returning control to the next window
        on the stack. If the stack is empty, the game will stop.
        This does return control, so remember to return immediately after
        calling it.
        """
        if len(self._windows) < 1:
            return
        handle('director.scene.exit')
        self._windows.pop()
        self._switch_window()
        if self._windows:
            handle('director.scene.enter')
        else:
            self.stop()
        # Empty all events!
        pygame.event.get()

    def push(self, window):
        """
        Place *window* on the top of the stack, and move control to it. This does
        return control, so remember to return immediately after calling it.

        :param window: The new window.
        :type window: :class:`Window <designer.core.window.Window>`
        """
        if self._windows:
            handle('director.scene.exit')
            self._switch_window()
        self._windows.append(window)
        window._register_default_events()
        handle('director.scene.enter')
        # Empty all events!
        pygame.event.get()

    def start(self, initial_game_state):
        """
        Starts Pygame main game loop. Checks for events and DirtySprite updates. Handles animations.

        :return: None
        """
        if not self._windows:
            return
        old_window = None
        window = self.current_window
        clock = window.clock
        stack = self._windows
        # Load up initial game state
        new_game_state = window._handle_event('director.start', Event(window=self))
        if new_game_state is not None:
            self._game_state = new_game_state
        else:
            self._game_state = initial_game_state
        del new_game_state
        del initial_game_state
        # Hide any unused references
        from designer.utilities.search import _detect_objects_recursively
        self._all_sprites = _detect_objects_recursively(self._game_state)
        for object in self._all_sprites:
            if object:
                object._reactivate()
        # Start running the game!
        self.running = True
        try:
            while self.running:
                window = stack[-1]
                if window is not old_window:
                    clock = window.clock
                    old_window = window

                    def frame_callback(interpolation):
                        """
                        A closure for handling drawing, which includes forcing the
                        rendering-related events to be fired.
                        """
                        window._handle_event("director.pre_render")
                        window._handle_event("director.render", Event(world=self._game_state))
                        window._draw()
                        window._handle_event("director.post_render")

                    def update_callback(delta):
                        """
                        A closure for handling events, which includes firing the update
                        related events (e.g., pre_update, update, and post_update).
                        """
                        if self.paused:
                            window._event_source.tick()
                            for event in window._event_source.get():
                                if event.type == pygame.QUIT:
                                    self.stop()
                            self._tick += 1
                            return
                        if len(pygame.event.get([pygame.VIDEOEXPOSE])) > 0:
                            window.redraw()
                            window._handle_event("director.redraw")

                        window._event_source.tick()
                        events = window._event_source.get()
                        for event in events:
                            window._queue_event(*_pygame_to_spyral(event, world=self._game_state))
                        window._handle_event("director.pre_update")
                        window._handle_event('director.update', Event(world=self._game_state, delta=delta))
                        self._tick += 1
                        window._handle_event("director.post_update")

                    clock.frame_callback = frame_callback
                    clock.update_callback = update_callback
                clock.tick()
        except GameEndException:
            pass

    def stop(self):
        """ Cleanly quits the game. """
        self.running = False
        pygame.quit()
        raise GameEndException("The game has ended correctly")

    def pause(self, new_state=True):
        self.paused = new_state



import math

from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
import pygame
import designer
import random
from designer.utilities.animation import *
from designer.utilities.easings import Linear, Iterate, Random


def glide_around(obj, speed):
    """
    Moves object around at random.

    :param objects: Object to move around
    :type objects: DesignerObject or DesignerGroup
    :param speed: Pixels to move per second
    :type speed: int
    :return: None
    """
    obj.animate(Animation('x', Random(-speed, speed), 1, absolute=False, loop=True))
    obj.animate(Animation('y', Random(-speed, speed), 1, absolute=False, loop=True))


def glide_right(obj, speed):
    """
    Moves object to the right of the window.

    :param obj: Object to move
    :type obj: DesignerObject or DesignerGroup
    :param speed: Pixels to move per second
    :type speed: int
    :return: None
    """
    obj.animate(Animation('x', Linear(obj['x'], get_width()), get_width() / speed))


def glide_left(obj, speed):
    """
    Moves object to the left of the window.

    :param obj: Object to move
    :type obj: DesignerObject
    :param speed: Pixels to move per second
    :type speed: int
    :return: None
    """
    obj.animate(Animation('x', Linear(obj['x'], 0), obj['x'] / speed))


def glide_up(obj, speed):
    """
    Moves object up on the window.

    :param obj: Object to move
    :type obj: DesignerObject or DesignerGroup
    :param speed: Pixels to move per second
    :type speed: int
    :return: None
    """
    obj.animate(Animation('y', Linear(obj['y'], 0), obj['y']/speed))


def glide_down(obj, speed):
    """
    Moves object down on the window.

    :param obj: Object to move
    :type obj: DesignerObject or DesignerGroup
    :param speed: Pixels to move per second
    :type speed: int
    :return: None
    """
    obj.animate(Animation('y', Linear(obj['y'], get_height()), (get_height()-obj['y'])/speed))


def glide_in_degrees(obj, direction, speed):
    """
    Moves object a given number of degrees in that direction

    :param obj: Object to move
    :type obj: DesignerObject or DesignerGroup
    :param direction: Direction in degrees counterclockwise for object to move
    :type direction: int
    :param speed: Pixels to move per second
    :type speed: int
    :return: None
    """
    raise NotImplementedError("This function is not ready yet!")


def spin(obj, duration=5, angle_limit=360):
    """
    Rotates object in place for a given number of degrees

    :param obj: Object to move
    :type obj: DesignerObject or DesignerGroup
    :param angle_limit: Degrees to rotate object
    :type angle_limit: int
    :param speed: Pixels to move per second
    :type speed: int
    :return: None
    """
    if isinstance(obj, DesignerObject):
        obj.animate(Animation('angle', Linear(0, angle_limit), duration, loop=True))


def linear_animation(obj, property, start, end, duration, absolute=False, shift=False, loop=False):
    obj.animate(Animation(property, Linear(start, end), duration, absolute, shift, loop))


def sequence_animation(obj, property, items, duration, times=1, absolute=False, shift=False, loop=False):
    obj.animate(Animation(property, Iterate(items, times), duration, absolute, shift, loop))



from designer.objects.designer_object import DesignerObject


def _detect_objects_recursively(data):
    """ Recursively detects any references to DesignerObjects,
     and returns them as a flattened set. """
    if isinstance(data, (list, set, tuple, frozenset)):
        result = set()
        for item in data:
            result.update(_detect_objects_recursively(item))
        return result
    elif isinstance(data, dict):
        result = set()
        for key, value in data.items():
            result.update(_detect_objects_recursively(key))
            result.update(_detect_objects_recursively(value))
        return result
    elif isinstance(data, DesignerObject):
        return {data}
    else:
        return set()



from typing import Optional

import pygame
import time
import operator
import inspect
import sys

import designer
from itertools import chain

from designer.core.event import COMMON_EVENT_NAME_LOOKUP, get_positional_event_parameters
from designer.core.internal_image import InternalImage
from designer.utilities.layer_tree import _LayerTree
from collections import defaultdict
from designer.core.clock import GameClock


def _has_value(obj, collect):
    for item in collect:
        if obj is item:
            return True
        elif isinstance(item, dict) and obj in item.values():
            return True
        elif isinstance(item, tuple) and obj in item:
            return True
    return False


class Window:
    """
    Creates a new Window. When a window is not active, no events will be processed
    for it. Windows are the basic units that are executed by Designer for your game,
    and can be subclassed and filled in with code which is relevant to your
    game. The :class:`Director <designer.core.designer.Director>`, is a manager for Windows,
    which maintains a stack and actually executes their code. There is always at least
    one default window, unless the game has ended.

    :param size: The `size` of the window internally (or "virtually"). This is
                 the coordinate space that you place Sprites in, but it does
                 not have to match up 1:1 to the window (which could be scaled).
    :type size: width, height
    :param int max_ups: Maximum updates to process per second. By default,
                        `max_ups` is pulled from the director.
    :param int max_fps: Maximum frames to draw per second. By default,
                        `max_fps` is pulled from the director.
    """

    def __init__(self, size=None, max_ups=None, max_fps=None):
        time_source = time.time
        self.clock = GameClock(
            time_source=time_source,
            max_fps=max_fps or designer.GLOBAL_DIRECTOR.fps,
            max_ups=max_ups or designer.GLOBAL_DIRECTOR.fps)
        self.clock.use_wait = True

        self._handlers = defaultdict(lambda: [])
        self._namespaces = set()
        self._event_source = designer.core.event.LiveEventHandler()
        self._handling_events = False
        self._events = []
        self._pending = []
        self._debug = True

        self._scale = designer.utilities.vector.Vec2D(1.0, 1.0)  # None
        self._surface = pygame.display.get_surface()
        display_size = self._surface.get_size()
        if size is not None:
            self.size = size
        else:
            self.size = display_size
        self._background: Optional[InternalImage] = None
        self._background_image = InternalImage(size=display_size)
        self._background_image.fill(designer.GLOBAL_DIRECTOR._window_color)
        self._background_version = 0
        self._surface.blit(self._background_image._surf, (0, 0))

        self._blits = []
        self._dirty_rects = []
        self._clear_this_frame = []
        self._clear_next_frame = []
        self._soft_clear = []
        self._static_blits = {}
        self._invalidating_views = {}
        self._collision_boxes = {}
        self._rect = self._surface.get_rect()

        self._layers = []
        self._child_views = []
        self._layer_tree = _LayerTree(self)
        self._objects = set()

        # View interface
        self._window = self
        self._views = []

        self._events_activated = False

    def _register_default_events(self):
        if not self._events_activated:
            designer.core.event.register('director.window.enter', self.redraw)
            designer.core.event.register('director.update', self._handle_events)
            designer.core.event.register('system.focus_change', self.redraw)
            designer.core.event.register('system.video_resize', self.redraw)
            designer.core.event.register('system.video_expose', self.redraw)
            designer.core.event.register('designer.internal.view.changed', self._invalidate_views)
            self._events_activated = True

    # Event Handling
    def _queue_event(self, event_name, event=None):
        """
        Internal method to add a new `event` to be handled by this window.

        :param str event_name: The name of the event to queue
        :param event: Metadata about this event.
        :type event: :class:`Event <designer.core.event.Event>`
        """
        if self._handling_events:
            self._pending.append((event_name, event))
        else:
            self._events.append((event_name, event))

    def _reg_internal(self, namespace, handlers, args,
                      kwargs, priority, dynamic):
        """
        Convenience method for registering a new event; other variations
        exist to keep the signature convenient and easy.
        """
        if namespace.endswith(".*"):
            namespace = namespace[:-2]
        self._namespaces.add(namespace)
        for handler in handlers:
            self._handlers[namespace].append((handler, args, kwargs,
                                              priority, dynamic))
        self._handlers[namespace].sort(key=lambda item: item[3]) # operator.itemgetter(3))

    def _get_namespaces(self, namespace):
        """
        Internal method for returning all the registered namespaces that are in
        the given namespace.
        """
        return [n for n in self._namespaces if (namespace == n or
                                                n.rsplit(".", 1)[0].startswith(namespace) or
                                                namespace.rsplit(".", 1)[0].startswith(n))]

    def _send_event_to_handler(self, event, event_type, handler: callable, args,
                               kwargs, priority, dynamic):
        """
        Internal method to dispatch events to their handlers.
        """
        def _get_arg_val(arg, default=inspect.Parameter.empty, index=None):
            if arg == 'event':
                return event
            elif hasattr(event, arg):
                return getattr(event, arg)
            else:
                if default != inspect.Parameter.empty:
                    return default
                positional_parameters = get_positional_event_parameters(event_type, event)
                if index < len(positional_parameters):
                    return getattr(event, positional_parameters[index])
                friendly_event_name = COMMON_EVENT_NAME_LOOKUP.get(event_type, event_type)
                suggestions = ", ".join(map(repr, [k for k in dir(event) if not k.startswith("__")]))
                raise TypeError(f"Your event handler function expected a parameter named {arg!r}, but "
                                f"the event {friendly_event_name!r} does not provide an argument with that name. "
                                f"The parameters allowed for this event are: {suggestions}")

        if dynamic is True:
            original_handler = handler
            handler = self
            for piece in original_handler.split("."):
                handler = getattr(handler, piece, None)
                if handler is None:
                    return
        if handler is sys.exit and args is None and kwargs is None:
            # Dirty hack to deal with python builtins
            args = []
            kwargs = {}
        elif args is None and kwargs is None:
            # Autodetect the arguments
            try:
                funct = handler.func
            except AttributeError as e:
                funct = handler
            try:
                signature = inspect.signature(funct)
            except Exception as e:
                return handler()
                #raise Exception(("Unfortunate Python Problem! "
                #                 f"{handler} isn't supported by Python's "
                #                 "inspect module! Oops."))
            h_args = [p.name for p in signature.parameters.values()]
            h_defaults = [p.default for p in signature.parameters.values()] or tuple()
            if len(h_args) > 0 and 'self' == h_args[0]:
                h_args.pop(0)
            d = len(h_args) - len(h_defaults)
            if d > 0:
                h_defaults = [inspect.Parameter.empty] * d + list(*h_defaults)
            args = [_get_arg_val(arg, default, index) for index, (arg, default)
                    in enumerate(zip(h_args, h_defaults))]
            kwargs = {}
        elif args is None:
            args = []
            kwargs = dict([(arg, _get_arg_val(arg)) for arg in kwargs])
        else:
            args = [_get_arg_val(arg) for arg in args]
            kwargs = {}
        if handler is not None:
            return handler(*args, **kwargs)

    def _handle_event(self, event_type, event=None, collect_results=False):
        """
        For a given event, send the event information to all registered handlers
        """
        handlers = chain.from_iterable(self._handlers[namespace]
                                       for namespace
                                       in self._get_namespaces(event_type))
        result = [] if collect_results else None
        for handler_info in handlers:
            new_result = self._send_event_to_handler(event, event_type, *handler_info)
            if new_result is not None:
                if collect_results:
                    result.append(new_result)
                else:
                    result = new_result
            if result is False:
                return result
        return result

    def _handle_events(self):
        """
        Run through all the events and handle them.
        """
        self._handling_events = True
        do = True
        while do or len(self._pending) > 0:
            do = False
            for (type, event) in self._events:
                self._handle_event(type, event)
            self._events = self._pending
            self._pending = []

    def _unregister_object_events(self, object):
        for name, handlers in self._handlers.items():
            self._handlers[name] = [h for h in handlers
                                    if not hasattr(h[0], '__self__') or h[0].__self__ is not object]
            if not self._handlers[name]:
                del self._handlers[name]

    def _unregister(self, event_namespace, handler):
        """
        Unregisters a registered handler for that namespace. Dynamic handler
        strings are supported as well. For more information, see
        `Event Namespaces`_.

        :param str event_namespace: An event namespace
        :param handler: The handler to unregister.
        :type handler: a function or string.
        """
        if event_namespace.endswith(".*"):
            event_namespace = event_namespace[:-2]
        self._handlers[event_namespace] = [h for h
                                           in self._handlers[event_namespace]
                                           if ((not hasattr(h[0], '__self__') and handler != h[0])
                                               or hasattr(h[0], '__self__')
                                                   and ((h[0].__func__ is not handler.__func__)
                                                        or (h[0].__self__ is not handler.__self__)))]
        if not self._handlers[event_namespace]:
            del self._handlers[event_namespace]

    def _clear_namespace(self, namespace):
        """
        Clears all handlers from namespaces that are at least as specific as the
        provided `namespace`. For more information, see `Event Namespaces`_.

        :param str namespace: The complete namespace.
        """
        if namespace.endswith(".*"):
            namespace = namespace[:-2]
        ns = [n for n in self._namespaces if n.startswith(namespace)]
        for namespace in ns:
            del self._handlers[namespace]

    def _clear_all_events(self):
        """
        Completely clear all registered events for this window. This is a very
        dangerous function, and should almost never be used.
        """
        self._handlers.clear()

    def _get_event_source(self):
        """
        The event source can be used to control event playback. Although
        normally events are given through the operating system, you can enforce
        events being played from a file; there is also a mechanism for recording
        events to a file.
        """
        return self._event_source

    def _set_event_source(self, source):
        self._event_source = source

    # Rendering
    @property
    def size(self):
        """
        Read-only property that returns a :class:`Vec2D <designer.utilities.vector.Vec2D>` of the
        width and height of the Window's size.  This is the coordinate space that
        you place Sprites in, but it does not have to match up 1:1 to the window
        (which could be scaled). This property can only be set once.
        """
        return self._size

    @size.setter
    def size(self, value):
        # This can only be called once.
        rsize = self._surface.get_size()
        self._size = value
        self._scale = (rsize[0] / value[0],
                       rsize[1] / value[1])

    @property
    def width(self):
        """
        The width of this window. Read-only number.
        """
        return self.size[0]

    @property
    def height(self):
        """
        The height of this window. Read-only number.
        """
        return self.size[1]

    @property
    def rect(self):
        """
        Returns a :class:`Rect <spyral.Rect>` representing the position (0, 0)
        and size of this Window.
        """
        return designer.utilities.rect.Rect((0, 0), self.size)

    @property
    def window(self):
        """
        Returns this window. Read-only.
        """
        return self._window

    @property
    def parent(self):
        """
        Returns this window. Read-only.
        """
        return self._window

    @property
    def background(self):
        """
        The background of this window. The given :class:`InternalImage <designer.core.internal_image.InternalImage>`
        must be the same size as the Window. A background will be handled
        intelligently by Spyral; it knows to only redraw portions of it rather
        than the whole thing, unlike a Sprite.
        """
        return self._background_image

    @background.setter
    def background(self, image: InternalImage):
        if self.size != image.size:
            image.scale(self.size)
        self._background_image = image
        self._background_version = image._version
        surface = image._surf
        size = self._surface.get_size()
        self._background = pygame.transform.smoothscale(surface, size)
        self._clear_this_frame.append(self._background.get_rect())


    def _register_object(self, object):
        """
        Internal method to add this object to the window
        """
        self._objects.add(object)
        # Add the view and its parents to the invalidating_views for the object
        parent_view = object._parent
        while parent_view != self:
            if parent_view not in self._invalidating_views:
                self._invalidating_views[parent_view] = set()
            self._invalidating_views[parent_view].add(object)
            parent_view = parent_view.parent

    def _unregister_object(self, object):
        """
        Internal method to remove this object from the window
        """
        if object in self._objects:
            self._objects.remove(object)
        if object in self._collision_boxes:
            del self._collision_boxes[object]
        for view in self._invalidating_views.keys():
            self._invalidating_views[view].discard(object)
        self._unregister_object_events(object)

    def _destroy_view(self, view):
        """
        Remove all references to the view from within this Window.
        """
        if view in self._invalidating_views:
            del self._invalidating_views[view]
        if view in self._collision_boxes:
            del self._collision_boxes[view]
        self._layer_tree.remove_view(view)

    def _blit(self, blit):
        """
        Apply any scaling associated with the Window to the Blit, then finalize
        it. Note that Window's don't apply cropping.
        """
        blit.apply_scale(self._scale)
        blit.finalize()
        self._blits.append(blit)

    def _static_blit(self, key, blit):
        """
        Identifies that this object will be statically blit from now, and
        applies scaling and finalization to the blit.
        """
        blit.apply_scale(self._scale)
        blit.finalize()
        self._static_blits[key] = blit
        #weakref.finalize(key, self._remove_static_blit, key)
        #weakref.finalize(key, print, "YOU KILLED", key)
        self._clear_this_frame.append(blit.rect)

    def _invalidate_views(self, view):
        """
        Expire any objects that belong to the view being invalidated.
        """
        if view in self._invalidating_views:
            for object in self._invalidating_views[view]:
                object._expire_static()

    def _remove_static_blit(self, key):
        """
        Removes this object from the static blit list
        """
        try:
            x = self._static_blits.pop(key)
            self._clear_this_frame.append(x.rect)
        except:
            pass

    def _draw(self):
        """
        Internal method that is called by the
        :class:`Director <designer.core.director.Director>` at the end of every .render() call
        to do the actual drawing.
        """

        # This function sits in a potential hot loop
        # For that reason, some . lookups are optimized away
        screen = self._surface

        # First we test if the background has been updated
        if self._background_version != self._background_image._version:
            self.background = self._background_image

        # Let's finish up any rendering from the previous frame
        # First, we put the background over all blits
        x = self._background.get_rect()
        for i in self._clear_this_frame + self._soft_clear:
            i = x.clip(i)
            b = self._background.subsurface(i)
            screen.blit(b, i)

        # Now, we need to blit layers, while simultaneously re-blitting
        # any static blits which were obscured
        static_blits = len(self._static_blits)
        dynamic_blits = len(self._blits)
        blits = list(self._static_blits.values()) + self._blits
        blits.sort(key=lambda l: (l.layer, l.id))

        # Clear this is a list of things which need to be cleared
        # on this frame and marked dirty on the next
        clear_this = self._clear_this_frame
        # Clear next is a list which will become clear_this on the next
        # draw cycle. We use this for non-static blits to say to clear
        # That spot on the next frame
        clear_next = self._clear_next_frame
        # Soft clear is a list of things which need to be cleared on
        # this frame, but unlike clear_this, they won't be cleared
        # on future frames. We use soft_clear to make things static
        # as they are drawn and then no longer cleared
        soft_clear = self._soft_clear
        self._soft_clear = []
        screen_rect = screen.get_rect()
        drawn_static = 0

        for blit in blits:
            blit_rect = blit.rect
            blit_flags = blit.flags
            # If a blit is entirely off screen, we can ignore it altogether
            if not screen_rect.contains(blit_rect) and not screen_rect.colliderect(blit_rect):
                continue
            # If this is a static blit...
            if blit.static:
                skip_soft_clear = False
                for rect in clear_this:
                    if blit_rect.colliderect(rect):
                        # One of the rects (needing to be cleared this frame and marked dirty on the next)
                        # is colliding with the current static blit's rect
                        # so we blit this static blit onto the screen and then add this static blit to the
                        # _soft_clear for next time
                        screen.blit(blit.surface, blit_rect, None, blit_flags)
                        skip_soft_clear = True
                        clear_this.append(blit_rect)
                        self._soft_clear.append(blit_rect)
                        drawn_static += 1
                        break
                if skip_soft_clear:
                    continue
                for rect in soft_clear:
                    if blit_rect.colliderect(rect):
                        screen.blit(blit.surface, blit_rect, None, blit_flags)
                        soft_clear.append(blit.rect)
                        drawn_static += 1
                        break
            else:
                if screen_rect.contains(blit_rect):
                    r = screen.blit(blit.surface, blit_rect, None, blit_flags)
                    clear_next.append(r)
                elif screen_rect.colliderect(blit_rect):
                    # Todo: See if this is ever called. Shouldn't be.
                    x = blit.rect.clip(screen_rect)
                    y = x.move(-blit_rect.left, -blit_rect.top)
                    b = blit.surface.subsurface(y)
                    r = screen.blit(blit.surface, blit_rect, None, blit_flags)
                    clear_next.append(r)

        if designer.GLOBAL_DIRECTOR.window_title is None:
            pygame.display.set_caption("%d / %d static, %d dynamic. %d ups, %d fps" %
                                      (drawn_static, static_blits,
                                       dynamic_blits, self.clock.ups,
                                       self.clock.fps))
        # Do the display update
        pygame.display.update(self._clear_next_frame + self._clear_this_frame)
        # Get ready for the next call
        self._clear_this_frame = self._clear_next_frame
        self._clear_next_frame = []
        self._blits = []

    def redraw(self):
        """
        Force the entire visible window to be completely redrawn.
        """
        self._clear_this_frame.append(pygame.Rect(self._rect))

    def _get_layer_position(self, view, layer):
        """
        For the given view and layer, calculate its position in the depth order.
        """
        return self._layer_tree.get_layer_position(view, layer)

    def _set_view_layer(self, view, layer):
        """
        Set the layer that the view is on within layer tree.
        """
        self._layer_tree.set_view_layer(view, layer)

    def _set_view_layers(self, view, layers):
        """
        Set the view's layers within the layer tree.
        """
        self._layer_tree.set_view_layers(view, layers)

    def _add_view(self, view):
        """
        Register the given view within this window.
        """
        self._layer_tree.add_view(view)

    @property
    def layers(self):
        """
        A list of strings representing the layers that are available for this
        window. The first layer is at the bottom, and the last is at the top.

        Note that the layers can only be set once.
        """
        return self._layers

    @layers.setter
    def layers(self, value):
        """
        Potential caveat: If you change layers after blitting, previous blits
        may be wrong for a frame, static ones wrong until they expire
        """
        if self._layers == []:
            self._layer_tree.set_view_layers(self, value)
            self._layers = value
        elif self._layers == value:
            pass
        else:
            raise Exception("You can only define the layers for a window once.")

    def _add_child(self, entity):
        """
        Add this child to the Window; since only Views care about their children,
        this function does nothing.
        """
        pass

    def _remove_child(self, entity):
        """
        Remove this child to the Window; since only Views care about their
        children, this function does nothing.
        """
        pass

    def _warp_collision_box(self, box):
        """
        Finalize the collision box. Don't apply scaling, because that's only
        for external rendering purposes.
        """
        box.finalize()
        return box

    def _set_collision_box(self, entity, box):
        """
        Registers the given entity (a View or Sprite) with the given
        CollisionBox.
        """
        self._collision_boxes[entity] = box

    def collide_objects(self, first, second):
        """
        Returns whether the first object is colliding with the second.

        :param first: A object or view
        :type first: :class:`DesignerObject <designer.objects.designer_object.DesignerObject>` or a
                     :class:`View <spyral.View>`
        :param second: Another object or view
        :type second: :class:`DesignerObject <designer.objects.designer_object.DesignerObject>` or a
                      :class:`View <spyral.View>`
        :returns: A ``bool``
        """
        if first not in self._collision_boxes or second not in self._collision_boxes:
            return False
        first_box = self._collision_boxes[first]
        second_box = self._collision_boxes[second]
        return first_box.collide_rect(second_box)

    def collide_point(self, obj, *args):
        """
        Returns whether the object is colliding with the point.

        :param obj: A DesignerObject
        :type obj: :class:`DesignerObject <designer.objects.designer_object.DesignerObject>`
        :param point: A point
        :type point: :class:`Vec2D <spyral.Vec2D>`
        :returns: A ``bool``
        """
        point = args
        if len(args) == 2:
            point = (args[0], args[1])
        elif len(args) != 1:
            raise ValueError(f"Incorrect number of arguments to collide_point: Expected x and y, got {args}")
        if obj not in self._collision_boxes:
            return False
        object_box = self._collision_boxes[obj]
        return object_box.collide_point(point)

    def collide_rect(self, obj, rect):
        """
        Returns whether the object is colliding with the rect.

        :param obj: A DesignerObject
        :type obj: :class:`DesignerObject <designer.objects.designer_object.DesignerObject>`
        :param rect: A rect
        :type rect: :class:`Rect <spyral.Rect>`
        :returns: A ``bool``
        """
        if obj not in self._collision_boxes:
            return False
        object_box = self._collision_boxes[obj]
        return object_box.collide_rect(rect)



"""GIFImage by Matthew Roe"""

from PIL import Image
import pygame

from designer.utilities.animation import Animation
from designer.utilities.easings import Iterate


class GifImage(Animation):
    def __init__(self, filename):
        self.filename = filename
        self.duration = .1
        self.total_duration = 0
        self.image = Image.open(filename)
        self.frames = []
        self.get_frames()

        # Handle per-frame durations
        print(len(self.frames))
        super().__init__('image', Iterate(self.frames), self.total_duration, True, None, True)

    def get_rect(self):
        return pygame.rect.Rect((0,0), self.image.size)

    def get_frames(self):
        image = self.image

        pal = image.getpalette()
        base_palette = []
        for i in range(0, len(pal), 3):
            rgb = pal[i:i+3]
            base_palette.append(rgb)

        all_tiles = []
        try:
            while 1:
                if not image.tile:
                    image.seek(0)
                if image.tile:
                    all_tiles.append(image.tile[0][3][0])
                image.seek(image.tell()+1)
        except EOFError:
            image.seek(0)

        all_tiles = tuple(set(all_tiles))

        try:
            while 1:
                try:
                    self.duration = image.info["duration"]
                except:
                    self.duration = 100

                self.duration *= .001 *10#convert to milliseconds!
                cons = False

                x0, y0, x1, y1 = (0, 0) + image.size
                if image.tile:
                    tile = image.tile
                else:
                    image.seek(0)
                    tile = image.tile
                if len(tile) > 0:
                    x0, y0, x1, y1 = tile[0][1]

                if all_tiles:
                    if all_tiles in ((6,), (7,)):
                        cons = True
                        pal = image.getpalette()
                        palette = []
                        for i in range(0, len(pal), 3):
                            rgb = pal[i:i+3]
                            palette.append(rgb)
                    elif all_tiles in ((7, 8), (8, 7)):
                        pal = image.getpalette()
                        palette = []
                        for i in range(0, len(pal), 3):
                            rgb = pal[i:i+3]
                            palette.append(rgb)
                    else:
                        palette = base_palette
                else:
                    palette = base_palette

                data = image.convert("RGBA").tobytes()
                pi = pygame.image.fromstring(data, image.size, "RGBA")#image.mode)
                #pi.set_palette(palette)
                if "transparency" in image.info:
                    pi.set_colorkey(image.info["transparency"])
                pi2 = pygame.Surface(image.size, pygame.SRCALPHA)
                if cons:
                    for i in self.frames:
                        pi2.blit(i[0], (0,0))
                #pi2.fill((255, 255, 0))
                pi2.blit(pi, (0, 0))
                #pi2.blit(pi, (x0, y0), (x0, y0, x1-x0, y1-y0))

                #self.frames.append([pi2, self.duration])
                self.total_duration += self.duration
                self.frames.append(pi2)
                image.seek(image.tell()+1)
        except EOFError:
            pass


    def get_height(self):
        return self.image.size[1]

    def get_width(self):
        return self.image.size[0]

    def get_size(self):
        return self.image.size

    def length(self):
        return len(self.frames)



import math

from designer.utilities.vector import Vec2D
from designer.mouse import get_mouse_position


def move_forward(object, amount, angle=None):
    if angle is None:
        angle = object.angle
    object.x += math.cos(math.radians(angle)) * amount
    object.y -= math.sin(math.radians(angle)) * amount
    return object


def move_backward(object, amount, angle=None):
    if angle is None:
        angle = object.angle
    object.x -= math.cos(math.radians(angle)) * amount
    object.y += math.sin(math.radians(angle)) * amount
    return object


def turn_right(object, amount):
    object.angle -= amount
    return object


def turn_left(object, amount):
    object.angle += amount
    return object


def go_to_xy(object, x, y):
    object.x = x
    object.y = y
    return object


def go_to(object, other_object):
    object.x = other_object.x
    object.y = other_object.y
    return object


def go_to_mouse(object):
    object.x, object.y = get_mouse_position()
    return object


"""
TODO: `glide` versions of all of these

"""


def point_in_direction(object, angle):
    object.angle = angle
    return object


def angle_between(p1, p2):
    # https://stackoverflow.com/a/64807404/1718155
    d1 = p2[0] - p1[0]
    d2 = p2[1] - p1[1]
    if d1 == 0:
        if d2 == 0:  # same points?
            deg = 0
        else:
            deg = 90 if p1[1] > p2[1] else 270
    elif d2 == 0:
        deg = 0 if p1[0] < p2[0] else 180
    else:
        deg = math.atan(d2 / d1) / math.pi * 180
        lowering = p1[1] < p2[1]
        if (lowering and deg < 0) or (not lowering and deg > 0):
            deg += 180
        else:
            deg += 0
    return deg


def point_towards(object, other_object):
    object.angle = angle_between(object.pos, other_object.pos)
    return object


def point_towards_mouse(object):
    object.angle = angle_between(object.pos, get_mouse_position())
    return object


def change_x(object, amount):
    object.x += amount
    return object


def change_y(object, amount):
    object.y += amount
    return object


def change_xy(object, x_amount, y_amount):
    object.x += x_amount
    object.y += y_amount
    return object


def set_x(object, new_x):
    object.x = new_x
    return new_x


def set_y(object, new_y):
    object.y = new_y
    return new_y


def get_x(object):
    return object.x


def get_y(object):
    return object.y


def get_angle(object):
    return object.angle


def set_scale(object, scale):
    object.scale = scale
    return object


def grow(object, times):
    object.scale = times
    return object


def shrink(object, times):
    object.scale = 1/times
    return object


def change_scale(object, amount):
    object.scale += amount
    return object


def get_scale(object):
    return object.scale


def grow_x(object, amount):
    object.scale_x += amount
    return object


def grow_y(object, amount):
    object.scale.y += amount
    return object


def set_scale_x(object, scale):
    object.scale_x = scale
    return object


def set_scale_y(object, scale):
    object.scale_y = scale
    return object


def get_scale_x(object):
    return object.scale_x


def get_scale_y(object):
    return object.scale_y


def show(object):
    object.visible = True
    return object


def hide(object):
    object.visible = False
    return object


def set_visible(object, status):
    object.visible = status
    return object


def get_visible(object):
    return object.visible


def flip_x(object):
    object.flip_x = not object.flip_x
    return object


def flip_y(object):
    object.flip_y = not object.flip_y
    return object


def set_flip_x(object, new_flip_x):
    object.flip_x = new_flip_x
    return object


def set_flip_y(object, new_flip_y):
    object.flip_y = new_flip_y
    return object


def get_flip_x(object):
    return object.flip_x


def get_flip_y(object):
    return object.flip_y


move_to_xy = go_to_xy
move_to_x = set_x
move_to_y = set_y
move_to = go_to
move_to_mouse = go_to_mouse



import difflib

import designer
from typing import List, Optional, Dict

from designer.core.window import Window
from designer.core.event import Event, register, unregister
from designer.utilities.vector import Vec2D
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities.rect import Rect
from designer.utilities.util import _anchor_offset, _Blit, _CollisionBox
from designer.utilities.animation import Animation


class DesignerObject:
    """
    DesignerObjects are how images and shapes are positioned and drawn onto the screen.
    They aggregate together information such as where to be drawn, their size, and their color.
    Each type of DesignerObject has its own specific kinds of attributes, but all DesignerObjects
    have certain common attributes.
    """
    FIELDS = (
        "rect", "pos",
        "layer",
        "x", "y", "width", "height",
        "size", "scale", "scale_x", "scale_y",
        "anchor",
        "angle", "flip_x", "flip_y", "visible",
        "parent", "mask",
        "alpha"
    )
    _ID = 0

    def __init__(self, parent=None, **kwargs):
        designer.check_initialized()

        if parent is None:
            parent = designer.GLOBAL_DIRECTOR.current_window

        # Unique ID for this object
        self._id = DesignerObject._ID
        DesignerObject._ID += 1

        #: Internal field about how long this DesignerObject has existed
        self._age: int = 0
        #: Whether or not this DesignerObject will not need to be redrawn for a while
        self._static: bool = False
        self._make_static = False

        # Independent Fields
        self._independent_fields = ('_pos', '_size', '_anchor', '_scale', '_angle', '_flip_x', '_flip_y')
        self._layer: Optional[str] = kwargs.get('layer', None)
        self._blend_flags = 0
        self._alpha = kwargs.get('alpha', 1.0)
        self._visible = kwargs.get('visible', True)
        self._pos = Vec2D(kwargs.get('pos')) if 'pos' in kwargs else Vec2D(kwargs.get('x', 0), kwargs.get('y', 0))
        self._size = Vec2D(kwargs.get('size')) if 'pos' in kwargs else Vec2D(kwargs.get('width', 1), kwargs.get('height', 1))
        self._anchor = kwargs.get('anchor', 'topleft')
        if 'scale' in kwargs and kwargs['scale'] != None:
            s = kwargs.get('scale')
            self._scale = Vec2D((s, s) if isinstance(s, (int, float)) else s)
        else:
            self._scale = Vec2D(kwargs.get('scale_x', 1.0), kwargs.get('scale_y', 1.0))
        self._angle = kwargs.get('angle', 0)
        self._flip_x = kwargs.get('flip_x', False)
        self._flip_y = kwargs.get('flip_y', False)
        self._active = kwargs.get('active', False)
        # TODO: Finish setting up cropping
        self._crop: Optional[Rect] = None
        self._mask: Optional[Rect] = None

        # Dependent fields
        self._offset = Vec2D(0, 0)
        self._computed_layer = 1
        #: The actual image after it has been scaled/cropped/rotated/etc.
        self._transform_image: Optional[DesignerSurface] = None
        self._transform_offset = Vec2D(0, 0)

        # Animation stuff
        self._animations: List[Animation] = []
        self._progress: Dict[Animation, float] = {}

        # Internal references to parents
        self._parent = parent
        self._window: Window = parent.window

        # Add it to the world
        if designer.GLOBAL_DIRECTOR.running:
            self._reactivate()

        self.FIELDS = set(self.FIELDS)

    def check_key(self, item):
        if item not in dir(self):
            suggestions = ", ".join(map(repr, difflib.get_close_matches(repr(item), self.FIELDS)))
            if suggestions:
                raise KeyError(f"Key {item!r} not found. Perhaps you meant one of these? {suggestions}")
            else:
                raise KeyError(
                    f"Key {item!r} not found. I didn't recognize that key, you should check the documentation!")

    def __getitem__(self, item):
        """ Allow this object to be treated like a dictionary. """
        self.check_key(item)
        return getattr(self, item)

    def __setitem__(self, key, value):
        """ Allow this object to be treated like a dictionary. """
        self.FIELDS.add(key)
        self.__setattr__(key, value)

    def _set_static(self):
        """
        Forces this class to be static, indicating that it will not be redrawn
        every frame.
        """
        self._make_static = True
        self._static = True

    def __del__(self):
        if self._static:
            self._window._remove_static_blit(self)

    def _expire_static(self):
        """
        Force this class to no longer be static; it will be redrawn for a few
        frames, until it has sufficiently aged. This also triggers the collision
        box to be recomputed.

        :rtype: bool
        :returns: whether it was successful
        """
        # Expire static is part of the private API which must
        # be implemented by Sprites that wish to be static.
        if self._static:
            self._window._remove_static_blit(self)
        self._static = False
        self._age = 0
        self._set_collision_box()
        return True

    def _make_blank_surface(self):
        self._transform_image = DesignerSurface((1, 1))
        self._recalculate_offset()
        self._expire_static()

    def _default_redraw_transforms(self, target):
        if self._flip_x or self._flip_y:
            target.flip(self._flip_x, self._flip_y)
        # Scale
        if self._scale != (1.0, 1.0):
            new_size = self._scale * target.size
            new_size = (int(new_size[0]), int(new_size[1]))
            if 0 in new_size:
                self._transform_image = DesignerSurface((1, 1))
                self._recalculate_offset()
                self._expire_static()
                return
            target.scale(new_size)
        # Rotate
        if self._angle != 0:
            angle = self._angle % 360
            # old = Vec2D(target.rect.center)
            target.rotate(angle)
            # new = target.rect.center
            # self._transform_offset = old - new
        # Finish updates
        self._transform_image = target._surf
        self._recalculate_offset()
        self._expire_static()

    def _redraw_internal_image(self):
        pass

    def _recalculate_offset(self):
        """
        Recalculates this designer object's offset based on its position, transform
        offset, anchor, its image, and the image's scaling.
        """

    @property
    def rect(self):
        """
        Returns a :class:`Rect <designer.Rect>` representing the position and size
        of this Designer Object's image. Note that if you change a property of this rect
        that it will not actually update this object's properties:

        >>> my_object = DesignerObject()
        >>> my_object.rect.top = 10

        Does not adjust the y coordinate of `my_object`. Changing the rect will
        adjust the object however

        >>> my_object.rect = designer.utilities.rect.Rect(10, 10, 64, 64)
        """
        return Rect(self._pos, self._size)

    @rect.setter
    def rect(self, *rect):
        if len(rect) == 1:
            r = rect[0]
            self.x, self.y = r.x, r.y
            self.width, self.height = r.w, r.h
        elif len(rect) == 2:
            self.pos = rect[0]
            self.size = rect[1]
        elif len(rect) == 4:
            self.x, self.y, self.width, self.height = rect
        else:
            raise ValueError("Too few arguments for the Rect of the Designer Object. Must have 1, 2, or 4")

    @property
    def pos(self):
        """
        The position of a sprite in 2D coordinates, represented as a
        :class:`Vec2D <spyral.Vec2D>`
        """
        return self._pos

    @pos.setter
    def pos(self, value):
        if value == self._pos:
            return
        self._pos = Vec2D(value)
        self._expire_static()

    @property
    def layer(self):
        """
        String. The name of the layer this sprite belongs to. See
        :ref:`layering <ref.layering>` for more.
        """
        return self._layer

    @layer.setter
    def layer(self, value: str):
        if value == self._layer:
            return
        self._layer = value
        self._computed_layer = self._window()._get_layer_position(self._parent(), value)
        self._expire_static()

    @property
    def x(self) -> int:
        """
        The x coordinate of the object, which will remain synced with the
        position.
        """
        return self.pos[0]

    @x.setter
    def x(self, v: int):
        self.pos = (v, self.y)

    @property
    def y(self):
        """
        The y coordinate of the object, which will remain synced with the
        position. Number
        """
        return self.pos[1]

    @y.setter
    def y(self, v):
        self.pos = (self.x, v)

    @property
    def anchor(self):
        """
        Defines an :ref:`anchor point <ref.anchors>` where coordinates are relative to
        on the internal_image. String.
        """
        return self._anchor

    @anchor.setter
    def anchor(self, value):
        if value == self._anchor:
            return
        self._anchor = value
        self._recalculate_offset()
        self._expire_static()

    @property
    def width(self):
        """
        The width of the object after all transforms. Number.
        """
        return self._size[0]

    @width.setter
    def width(self, value):
        self.size = (value, self.height)

    @property
    def height(self):
        """
        The height of the image after all transforms. Number.
        """
        return self._size[1]

    @height.setter
    def height(self, value):
        self.size = (self.width, value)

    @property
    def size(self):
        """
        The size of the image after all transforms (:class:`Vec2D <designer.utilities.vector.Vec2D>`).
        """
        return self._size

    @size.setter
    def size(self, value):
        self._size = Vec2D(value)
        self._redraw_internal_image()
        self._expire_static()

    @property
    def scale(self):
        """
        A scale factor for resizing the image. When read, it will always contain
        a :class:`designer.utilities.vector.Vec2D` with an x factor and a y factor, but it can be
        set to a numeric value which wil ensure identical scaling along both
        axes.
        """
        return self._scale

    @scale.setter
    def scale(self, value):
        if isinstance(value, (int, float)):
            value = Vec2D(value, value)
        if self._scale == value:
            return
        self._scale = Vec2D(value)
        self._redraw_internal_image()
        self._expire_static()

    @property
    def scale_x(self):
        """
        The x factor of the scaling that's kept in sync with scale. Number.
        """
        return self._scale[0]

    @scale_x.setter
    def scale_x(self, x):
        self.scale = (x, self._scale[1])

    @property
    def scale_y(self):
        """
        The y factor of the scaling that's kept in sync with scale. Number.
        """
        return self._scale[1]

    @scale_y.setter
    def scale_y(self, y):
        self.scale = (self._scale[0], y)

    @property
    def angle(self):
        """
        An angle to rotate the image by. Rotation is computed after scaling and
        flipping, and keeps the center of the original image aligned with the
        center of the rotated image.
        """
        return self._angle

    @angle.setter
    def angle(self, value):
        if self._angle == value:
            return
        self._angle = value
        self._redraw_internal_image()

    @property
    def flip_x(self):
        """
        A boolean that determines whether the object should be flipped
        horizontally.
        """
        return self._flip_x

    @flip_x.setter
    def flip_x(self, value):
        if self._flip_x == value:
            return
        self._flip_x = value
        self._redraw_internal_image()

    @property
    def flip_y(self):
        """
        A boolean that determines whether the object should be flipped
        vertically.
        """
        return self._flip_y

    @flip_y.setter
    def flip_y(self, value):
        if self._flip_y == value:
            return
        self._flip_y = value
        self._redraw_internal_image()

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        if self._alpha == value:
            return
        self._alpha = value
        self._expire_static()

    @property
    def visible(self):
        """
        A boolean indicating whether this object should be drawn.
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if self._visible == value:
            return
        self._visible = value
        self._expire_static()

    @property
    def active(self) -> bool:
        """
        A boolean indicating whether the object is *active*, aka it should be drawn and have collisions, animations,
        and other events handled. An object is active when it is first created while the game is running, but inactive
        if it is created before the game is started.
        """
        return self._active

    @active.setter
    def active(self, value):
        if value:
            self.reactivate()
        else:
            self.destroy()

    @property
    def window(self):
        """
        The top-level window that this object belongs to. Read-only.
        """
        return self._window

    @property
    def parent(self):
        """
        The parent of this object, either a :class:`View <designer.objects.view.View>` or a
        :class:`Window <desinger.objects.view.View>`. Read-only.
        """
        return self._parent

    @property
    def mask(self):
        """
        A :class:`Rect <designer.utilities.rect.Rect>` to use instead of the current object's rect
        for computing collisions. `None` if the object's rect should be used.
        """
        return self._mask

    @mask.setter
    def mask(self, value):
        self._mask = value
        self._set_collision_box()

    def _draw(self):
        """
        Internal method for generating this object's blit, unless it is
        invisible or currently static. If it has aged sufficiently or is being
        forced, it will become static; otherwise, it ages one step.
        """
        if not self.visible:
            return
        if self._transform_image is None:
            return
        if self._static:
            return

        # TODO: Make sure this is sufficient
        self._transform_image.set_alpha(int(self._alpha * 255))

        area = Rect(self._transform_image.get_rect())
        b = _Blit(self._transform_image, self._pos - self._offset,
                  area, self._computed_layer, self._blend_flags, False,
                  self._id)

        if self._make_static or self._age > 4:
            b.static = True
            self._make_static = False
            self._static = True
            self._parent._static_blit(self, b)
            return
        self._parent._blit(b)
        self._age += 1

    def _set_collision_box(self):
        """
        Updates this object's collision box.
        """
        if self._transform_image is None:
            return
        if self._mask is None:
            area = Rect(self._transform_image.get_rect())
        else:
            area = self._mask
        c = _CollisionBox(self._pos - self._offset, area)
        warped_box = self._parent._warp_collision_box(c)
        self._window._set_collision_box(self, warped_box.rect)

    def destroy(self):
        """
        When you no longer need an Object, you can call this method to have it
        removed from the Window. This will not remove the object entirely from
        memory if there are other references to it; if you need to do that,
        remember to ``del`` the reference to it.
        """
        self._active = False
        designer.GLOBAL_DIRECTOR._untrack_object(self)
        self._window._unregister_object(self)
        self._window._remove_static_blit(self)
        self._parent._remove_child(self)
        unregister('director.render', self._draw)

    def _reactivate(self):
        """
        Internal method for making an Object active again.
        Not a preferred mechanism, may have undefined behavior.

        TODO: Finish this so that it can actually work if people want this.
        """
        self._active = True
        designer.GLOBAL_DIRECTOR._track_object(self)
        self.parent._add_child(self)
        self._window._register_object(self)
        self._age = 0
        self._static = False
        register('director.render', self._draw)

    # Animation Methods

    def _evaluate(self, animation, progress):
        """
        Performs a step of the given animation, setting any properties that will
        change as a result of the animation (e.g., x position).
        """
        values = animation.evaluate(self, progress)
        for a_property in animation.properties:
            if a_property in values:
                setattr(self, a_property, values[a_property])

    def _run_animations(self, delta):
        """
        For a given time-step (delta), perform a step of all the animations
        associated with this designer object.
        """
        completed = []
        for animation in self._animations:
            self._progress[animation] += delta
            progress = self._progress[animation]
            if progress > animation.duration:
                self._evaluate(animation, animation.duration)
                if animation.loop is True:
                    self._evaluate(animation, progress - animation.duration)
                    self._progress[animation] = progress - animation.duration
                elif animation.loop:
                    current = progress - animation.duration + animation.loop
                    self._evaluate(animation, current)
                    self._progress[animation] = current
                else:
                    completed.append(animation)
            else:
                self._evaluate(animation, progress)
        # Stop all completed animations
        for animation in completed:
            self.stop_animation(animation)

    def animate(self, animation):
        """
        Animates this object given an animation. Read more about
        :class:`animation <designer.animation>`.

        :param animation: The animation to run.
        :type animation: :class:`Animation <designer.Animation>`
        """
        for a in self._animations:
            repeats = a.properties.intersection(animation.properties)
            if repeats:
                # Loop over all repeats
                raise ValueError(f"Cannot animate on properties {repeats} twice")
        if len(self._animations) == 0:
            designer.core.event.register('director.update',
                                         self._run_animations,
                                         ('delta',))
        self._animations.append(animation)
        self._progress[animation] = 0
        self._evaluate(animation, 0.0)
        e = Event(animation=animation, sprite=self)
        # Loop over all possible properties
        for a_property in animation.properties:
            designer.core.event.handle(f"{self.__class__.__name__}.{a_property}.animation.start", e)

    def stop_animation(self, animation):
        """
        Stops a given animation currently running on this object.

        :param animation: The animation to stop.
        :type animation: :class:`Animation <spyral.Animation>`
        """
        if animation in self._animations:
            self._animations.remove(animation)
            del self._progress[animation]
            e = Event(animation=animation, sprite=self)
            for a_property in animation.properties:
                designer.core.event.handle(f"{self.__class__.__name__}.{a_property}.animation.end", e)
            if len(self._animations) == 0:
                designer.core.event.unregister('director.update', self._run_animations)

    def stop_all_animations(self):
        """
        Stops all animations currently running on this object.
        """
        for animation in self._animations:
            self.stop_animation(animation)

    def collide_other(self, other):
        """
        Returns whether this object is currently colliding with the other
        object. This collision will be computed correctly regarding the objects
        offsetting and scaling within their views.

        :param other: The other object
        :type other: :class:`DesignerObject <designer.objects.designer_object.DesignerObject>`
        :returns: ``bool`` indicating whether this object is colliding with the
                  other object.
        """
        return self._window.collide_objects(self, other)

    def collide_point(self, point):
        """
        Returns whether this object is currently colliding with the position.
        This uses the appropriate offsetting for the object within its views.

        :param point: The point (relative to the window dimensions).
        :type point: :class:`Vec2D <designer.utilities.vector.Vec2D>`
        :returns: ``bool`` indicating whether this object is colliding with the
                  position.
        """
        return self._window.collide_point(self, *point)

    def collide_rect(self, rect):
        """
        Returns whether this object is currently colliding with the rect. This
        uses the appropriate offsetting for the object within its views.

        :param rect: The rect (relative to the window dimensions).
        :type rect: :class:`Rect <designer.utilities.rect.Rect>`
        :returns: ``bool`` indicating whether this object is colliding with the
                  rect.
        """
        return self._window.collide_rect(self, rect)



"""The keyboard modules provides an interface to adjust the keyboard's repeat
rate.

.. attribute:: repeat

    When the keyboard repeat is enabled, keys that are held down will keep
    generating new events over time. Defaults to `False`.

.. attribute:: delay

    `int` to control how many milliseconds before the repeats start.

.. attribute:: interval

    `int` to control how many milliseconds to wait between repeated events.

"""
import pygame


class KeyboardModule:
    DEFAULT_DELAY = 600
    DEFAULT_REPEAT = False
    DEFAULT_INTERVAL = 100

    def __init__(self):
        self._repeat = False
        self._delay = 600
        self._interval = 100

    def _update_repeat_status(self):
        if self._repeat:
            pygame.key.set_repeat(self._delay, self._interval)
        else:
            pygame.key.set_repeat()

    @property
    def repeat(self):
        return self._repeat

    @repeat.setter
    def repeat(self, value):
        self._repeat = value
        self._update_repeat_status()

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, value):
        self._interval = value
        self._update_repeat_status()

    @property
    def delay(self):
        return self._delay

    @delay.setter
    def delay(self, value):
        self._delay = value
        if value == 0:
            self._repeat = False
        self._update_repeat_status()




"""
Rects are a convenience class for managing rectangular regions.
"""

import pygame
import designer


class Rect:
    """
    Rect represents a rectangle and provides some useful features. Rects can
    be specified 3 ways in the constructor:

    #. Four numbers, ``x``, ``y``, ``width``, ``height``
    #. Two tuples, ``(x, y)`` and `(width, height)`
    #. Another rect, which is copied

    >>> rect1 = designer.utilities.vector.Rect(10, 10, 64, 64)               # Method 1
    >>> rect2 = designer.utilities.vector.Rect((10, 10), (64, 64))           # Method 2
    >>> rect3 = designer.utilities.vector.Rect(rect1.topleft, rect1.size)    # Method 2
    >>> rect4 = designer.utilities.vector.Rect(rect3)                        # Method 3

    Rects support all the usual :ref:`anchor points <ref.anchors>` as
    attributes, so you can both get ``rect.center`` and assign to it.
    Rects also support attributes of ``right``, ``left``, ``top``, ``bottom``,
    ``x``, and ``y``.

    >>> rect1.x
    10
    >>> rect1.centerx
    42.0
    >>> rect1.width
    64
    >>> rect1.topleft
    Vec2D(10, 10)
    >>> rect1.bottomright
    Vec2D(74, 74)
    >>> rect1.center
    Vec2D(42.0, 42.0)
    >>> rect1.size
    Vec2D(64, 64)

    """

    _x: float
    _y: float
    _w: float
    _h: float

    def __init__(self, *args):
        if len(args) == 1:
            r = args[0]
            self._x, self._y = r.x, r.y
            self._w, self._h = r.w, r.h
        elif len(args) == 2:
            self._x, self._y = args[0]
            self._w, self._h = args[1]
        elif len(args) == 4:
            self.left, self.top, self.width, self.height = args
        else:
            raise ValueError("You done goofed.")

    def __getattr__(self, name):
        v = designer.utilities.vector.Vec2D
        if name == "right":
            return self._x + self._w
        if name == "left" or name == "x":
            return self._x
        if name == "top" or name == "y":
            return self._y
        if name == "bottom":
            return self._y + self._h
        if name == "topright":
            return v(self._x + self._w, self._y)
        if name == "bottomleft":
            return v(self._x, self._y + self._h)
        if name == "topleft" or name == "pos":
            return v(self._x, self._y)
        if name == "bottomright":
            return v(self._x + self._w, self._y + self._h)
        if name == "centerx":
            return self._x + self._w / 2.
        if name == "centery":
            return self._y + self._h / 2.
        if name == "center":
            return v(self._x + self._w / 2., self._y + self._h / 2.)
        if name == "midleft":
            return v(self._x, self._y + self._h / 2.)
        if name == "midright":
            return v(self._x + self._w, self._y + self._h / 2.)
        if name == "midtop":
            return v(self._x + self._w / 2., self._y)
        if name == "midbottom":
            return v(self._x + self._w / 2., self._y + self._h)
        if name == "size":
            return v(self._w, self._h)
        if name == "width" or name == "w":
            return self._w
        if name == "height" or name == "h":
            return self._h

        raise AttributeError("type object 'rect' "
                             "has no attribute '%s'" % name)

    def __setattr__(self, name, val):
        # This could use _a lot_ more error checking
        if name[0] == "_":
            self.__dict__[name] = int(val)
            return
        if name == "right":
            self._x = val - self._w
        elif name == "left":
            self._x = val
        elif name == "top":
            self._y = val
        elif name == "bottom":
            self._y = val - self._h
        elif name == "topleft" or name == "pos":
            self._x, self._y = val
        elif name == "topright":
            self._x = val[0] - self._w
            self._y = val[1]
        elif name == "bottomleft":
            self._x = val[0]
            self._y = val[1] - self._h
        elif name == "bottomright":
            self._x = val[0] - self._w
            self._y = val[0] - self._h
        elif name == "width" or name == "w":
            self._w = val
        elif name == "height" or name == "h":
            self._h = val
        elif name == "size":
            self._w, self._h = val
        elif name == "centerx":
            self._x = val - self._w / 2.
        elif name == "centery":
            self._y = val - self._h / 2.
        elif name == "center":
            self._x = val[0] - self._w / 2.
            self._y = val[1] - self._h / 2.
        elif name == "midtop":
            self._x = val[0] - self._w / 2.
            self._y = val[1]
        elif name == "midleft":
            self._x = val[0]
            self._y = val[1] - self._h / 2.
        elif name == "midbottom":
            self._x = val[0] - self._w / 2.
            self._y = val[1] - self._h
        elif name == "midright":
            self._x = val[0] - self._w
            self._y = val[1] - self._h / 2.
        else:
            raise AttributeError(f"The attribute {name} was not recognized.")

    def copy(self):
        """
        Returns a copy of this rect

        :returns: A new :class:`Rect <designer.utilities.vector.Rect>`
        """
        return Rect(self._x, self._y, self._w, self._h)

    def move(self, x, y):
        """
        Returns a copy of this rect offset by *x* and *y*.

        :param float x: The horizontal offset.
        :param float y: The vertical offset.
        :returns: A new :class:`Rect <designer.utilities.vector.Rect>`
        """
        return Rect(x, y, self._w, self._h)

    def move_ip(self, x, y):
        """
        Moves this rect by *x* and *y*.

        :param float x: The horizontal offset.
        :param float y: The vertical offset.
        """
        self._x, self._y = self._x + x, self._y + y

    def inflate(self, width, height):
        """
        Returns a copy of this rect inflated by *width* and *height*.

        :param float width: The amount to add horizontally.
        :param float height: The amount to add vertically.
        :returns: A new :class:`Rect <designer.utilities.vector.Rect>`
        """
        c = self.center
        n = self.copy()
        n.size = (self._w + width, self._h + height)
        n.center = c
        return n

    def inflate_ip(self, width, height):
        """
        Inflates this rect by *width*, *height*.

        :param float width: The amount to add horizontally.
        :param float height: The amount to add vertically.
        """
        c = self.center
        self.size = (self._w + width, self._h + height)
        self.center = c

    def union(self, other):
        """
        Returns a new rect which represents the union of this rect
        with other -- in other words, a new rect is created that can fit both
        original rects.

        :param other: The other Rect.
        :type other: :class:`Rect <designer.utilities.vector.Rect>`
        :returns: A new :class:`Rect <designer.utilities.vector.Rect>`
        """
        top = min(self.top, other.top)
        left = min(self.left, other.left)
        bottom = max(self.bottom, other.bottom)
        right = max(self.right, other.right)
        return Rect((left, top), (right - left, bottom - top))

    def union_ip(self, other):
        """
        Modifies this rect to be the union of it and the other -- in other
        words, this rect will expand to include the other rect.

        :param other: The other Rect.
        :type other: :class:`Rect <designer.utilities.vector.Rect>`
        """
        top = min(self.top, other.top)
        left = min(self.left, other.left)
        bottom = max(self.bottom, other.bottom)
        right = max(self.right, other.right)
        self.top, self.left = top, left
        self.bottom, self.right = bottom, right

    # @test: Rect(10,10,50,50).clip(Rect(0,0,20,20)) -> Rect(10,10,10,10)
    def clip(self, other):
        """
        Returns a Rect which is cropped to be completely inside of other.
        If the other does not overlap with this rect, a rect of size 0 is
        returned.

        :param other: The other Rect.
        :type other: :class:`Rect <designer.utilities.vector.Rect>`
        :returns: A new :class:`Rect <designer.utilities.vector.Rect>`
        """
        B = other
        A = self
        try:
            B._x
        except TypeError:
            B = Rect(B)

        if A._x >= B._x and A._x < (B._x + B._w):
            x = A._x
        elif B._x >= A._x and B._x < (A._x + A._w):
            x = B._x
        else:
            return Rect(A._x, A._y, 0, 0)

        if ((A._x + A._w) > B._x) and ((A._x + A._w) <= (B._x + B._w)):
            w = A._x + A._w - x
        elif ((B._x + B._w) > A._x) and ((B._x + B._w) <= (A._x + A._w)):
            w = B._x + B._w - x
        else:
            return Rect(A._x, A._y, 0, 0)

        if A._y >= B._y and A._y < (B._y + B._h):
            y = A._y
        elif B._y >= A._y and B._y < (A._y + A._h):
            y = B._y
        else:
            return Rect(A._x, A._y, 0, 0)

        if ((A._y + A._h) > B._y) and ((A._y + A._h) <= (B._y + B._h)):
            h = A._y + A._h - y
        elif ((B._y + B._h) > A._y) and ((B._y + B._h) <= (A._y + A._h)):
            h = B._y + B._h - y
        else:
            return Rect(A._x, A._y, 0, 0)

        return Rect(x, y, w, h)

    def clip_ip(self, other):
        """
        Modifies this rect to be cropped completely inside of other.
        If the other does not overlap with this rect, this rect will have a size
        of 0.

        :param other: The other Rect.
        :type other: :class:`Rect <designer.utilities.vector.Rect>`
        """
        new_rect = self.clip(other)
        self.topleft, self.size = new_rect.topleft, new_rect.size

    def contains(self, other):
        """
        Returns `True` if the other rect is contained inside this rect.

        :param other: The other Rect.
        :type other: :class:`Rect <designer.utilities.vector.Rect>`
        :returns: A `bool` indicating whether this rect is contained within
                  another.
        """
        return (other.collide_point(self.topleft) and
                other.collide_point(self.bottomright))

    def collide_rect(self, other):
        """
        Returns `True` if this rect collides with the other rect.

        :param other: The other Rect.
        :type other: :class:`Rect <designer.utilities.vector.Rect>`
        :returns: A `bool` indicating whether this rect is contained within
                  another.
        """
        return (self.clip(other).size != (0, 0) or
                other.clip(self).size != (0, 0))

    def collide_point(self, point):
        """
        :param point: The point.
        :type point: :class:`Vec2D <spyral.Vec2D>`
        :returns: A `bool` indicating whether the point is contained within this
                  rect.
        """
        # This could probably be optimized as well
        return point[0] > self.left and point[0] < self.right and \
               point[1] > self.top and point[1] < self.bottom

    def _to_pygame(self):
        """
        Internal method for creating a Pygame compatible rect from this rect.

        :returns: A :class:`pygame.Rect`
        """
        return pygame.Rect((self.left, self.top), (self.width, self.height))

    def __str__(self):
        return ''.join(['<rect(',
                        str(self._x),
                        ',',
                        str(self._y),
                        ',',
                        str(self._w),
                        ',',
                        str(self._h),
                        ')>'])

    def __repr__(self):
        return self.__str__()



"""Animations interpolate a property between two values over a number of frames.
They can be combined to run at the same time, or directly after each other."""
from designer.utilities.vector import Vec2D


class Animation:
    """
    Creates an animation on *property*, with the specified
    *easing*, to last *duration* in seconds.

    The following example shows a Sprite with an animation that will linearly
    change its 'x' property from 0 to 100 over 2 seconds.::

        from spyral import Sprite, Animation, easing
        ...
        my_sprite = Sprite(my_scene)
        my_animation = Animation('x', easing.Linear(0, 100), 2.0)
        my_sprite.animate(my_animation)

    Animations can be appended one after another with the `+`
    operator, and can be run in parallel with the `&` operator.

    >>> from designer import Animation, Linear
    >>> first  = Animation('x', Linear(0, 100), 2.0)
    >>> second = Animation('y', Linear(0, 100), 2.0)
    # Sequential animations
    >>> right_angle = first + second
    # Parallel animations
    >>> diagonal = first & second

    :param property: The property of the object to change (e.g., 'x')
    :type property: :class:`string`
    :param easing: The easing (rate of change) of the property.
    :type easing: :class:`Easing <spyral.Easing>`
    :param duration: How many seconds to play the animation
    :type duration: :class:`float`
    :param absolute: (**Unimplemented?**) Whether to position this relative
                     to the object's offset, or to absolutely position it on the
                     screen.
    :type absolute: :class:`boolean`
    :param shift: How much to offset the animation (a number if the property is
                  scalar, a :class:`Vec2D <designer.utilities.vector.Vec2D>` if the property is
                  "pos", and None if there is no offset.
    :type shift: None, a :class:`Vec2D <designer.utilities.vector.Vec2D>`, or a number
    :param loop: Whether to loop indefinitely
    :type loop: :class:`boolean`
    """

    def __init__(self, property,
                 easing,
                 duration=1.0,
                 absolute=True,
                 shift=None,
                 loop=False
                 ):
        # Idea: These easings could be used for camera control
        # at some point. Everything should work pretty much the same.
        self.absolute = absolute
        self.property = property
        self.easing = easing
        self.duration = duration
        self.loop = loop
        self.properties = {property}
        self._shift = shift

    def evaluate(self, object, progress):
        """
        For a given *object*, complete *progress*'s worth of this animation.
        Basically, complete a step of the animation. Returns a dictionary
        representing the changed property and its new value, e.g.:
        :code:`{"x": 100}`. Typically, you will use the object's animate function instead of calling
        this directly.

        :param object: The Sprite that will be manipulated.
        :type object: :class:`Sprite <spyral.Sprite>`
        :param float progress: The amount of progress to make on this animation.
        :rtype: :class:`dict`
        """
        progress = progress / self.duration
        value = self.easing(object, progress)
        original = getattr(object, self.property)
        if not self.absolute:
            if isinstance(original, (tuple, list, Vec2D)):
                value = (value[0] + original[0],
                         value[1] + original[1])
            else:
                value += original
        if self._shift is not None:
            if isinstance(original, (tuple, list, Vec2D)):
                value = (value[0] + self._shift[0],
                         value[1] + self._shift[1])
            else:
                value = value + self._shift
        return {self.property: value}

    def __and__(self, second):
        return MultiAnimation(self, second)

    def __iand__(self, second):
        return MultiAnimation(self, second)

    def __add__(self, second):
        return SequentialAnimation(self, second)

    def __iadd__(self, second):
        return SequentialAnimation(self, second)


class MultiAnimation(Animation):
    """
    Class for creating parallel animation from two other animations.

    This does not respect the absolute setting on individual
    animations. Pass absolute as a keyword argument to change,
    default is True.
    Absolute applies only to numerical properties.

    loop is accepted as a kwarg, default is True if any child
    loops, or False otherwise.
    """
    def __init__(self, *animations, **kwargs):
        self.properties = set()
        self._animations = []
        self.duration = 0
        self.absolute = kwargs.get('absolute', True)
        self.loop = False
        for animation in animations:
            i = animation.properties.intersection(self.properties)
            if i:
                message = "Cannot animate on the same properties twice: %s"
                raise ValueError(message % i)
            self.properties.update(animation.properties)
            self._animations.append(animation)
            self.duration = max(self.duration, animation.duration)
            if animation.loop:
                self.loop = True
        # Ensure we don't clobber on properties
        clobbering_animations = [('scale', {'scale_x', 'scale_y'}),
                                 ('pos', {'x', 'y', 'position'}),
                                 ('position', {'x', 'y', 'pos'})]
        for prop, others in clobbering_animations:
            overlapping_properties = self.properties.intersection(others)
            if prop in self.properties and overlapping_properties:
                message = "Cannot animate on %s and %s in the same animation."
                raise ValueError(message % (prop,
                                            overlapping_properties.pop()))
        self.loop = kwargs.get('loop', self.loop)

    def evaluate(self, object, progress):
        res = {}
        for animation in self._animations:
            if progress <= animation.duration:
                res.update(animation.evaluate(object, progress))
            else:
                res.update(animation.evaluate(object, animation.duration))
        return res


class SequentialAnimation(Animation):
    """
    An animation that represents the input animations in sequence.

    loop is accepted as a kwarg, default is False.

    If the last animation in a SequentialAnimation is set to loop,
    that animation will be looped indefinitely at the end, but not
    the entire SequentialAnimation. If loop is set to true, the
    entire SequentialAnimation will loop indefinitely.
    """
    def __init__(self, *animations: Animation, **kwargs):
        self.properties = set()
        self._animations = list(animations)
        self.duration = 0
        self.absolute = True
        self.loop = kwargs.get('loop', False)
        for animation in animations:
            self.properties.update(animation.properties)
            self.duration += animation.duration
            if self.loop and animation.loop:
                raise ValueError("Looping sequential animation with a looping "
                                 "animation anywhere in the sequence "
                                 "is not allowed.")
            if animation.loop and animation is not animations[-1]:
                raise ValueError("Looping animation in the middle of a "
                                 "sequence is not allowed.")
        if animations and animations[-1].loop is True:
            self.loop = self.duration - animations[-1].duration

    def evaluate(self, object, progress):
        res = {}
        # We have no animations
        if not self._animations:
            return res
        # We reached the end
        if progress == self.duration:
            res.update(self._animations[-1].evaluate(object,
                       self._animations[-1].duration))
            return res
        # Progress through all the animations till we reach our current
        i = 0
        while progress > self._animations[i].duration:
            progress -= self._animations[i].duration
            i += 1
        # As long as we reached one, let's evaluate the previous?
        if i > 0:
            res.update(self._animations[i - 1].evaluate(object, self._animations[i - 1].duration))
        res.update(self._animations[i].evaluate(object, progress))
        return res

    def append(self, another):
        self.properties.update(another.properties)
        self.duration += another.duration
        if another.loop is True and (not self._animations or self._animations[-1].loop is False):
            self.loop = self.duration - another.duration
        self._animations.append(another)


class DelayAnimation(Animation):
    """
    Animation which performs no actions. Useful for lining up appended
    and parallel animations so that things run at the right times.
    """
    def __init__(self, duration=1.0):
        super().__init__(duration=duration, absolute=False, properties=set(), loop=False)

    def evaluate(self, object, progress):
        return {}



import pygame

import math

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities.vector import Vec2D
from designer.utilities.util import _anchor_offset


class Arc(DesignerObject):
    DEFAULT_BORDER_WIDTH = 1
    FIELDS = (*DesignerObject.FIELDS, 'color', 'border')

    def __init__(self, center, width, height, start_angle, stop_angle, anchor, color, border, **kwargs):
        """
        Creates an Arc Designer Object on window.

        :param color: color to draw arc
        :type color: str or List[str]
        :param start_angle: angle to start drawing arc at
        :type start_angle: int
        :param stop_angle: angle to stop drawing arc at
        :type stop_angle: int
        :param thickness: thickness of arc in pixels
        :type thickness: int
        :param left: x coordinate of top left corner of arc
        :type left: int
        :param top: y coordinate of top left corner of arc
        :type top: int
        :param width: width of arc to be drawn
        :type width: int
        :param height: height of arc to be drawn
        :type height: int
        """
        super().__init__(**kwargs)

        x, y = center
        x = x if x is not None else get_width() / 2
        y = y if y is not None else get_height() / 2
        center = x, y

        self._pos = center
        self._anchor = anchor
        # Ellipse specific data
        self._size = Vec2D(width, height)
        self._start_angle = start_angle
        self._stop_angle = stop_angle
        self._color = color
        self._border = border

        # And draw!
        self._redraw_internal_image()

    def _recalculate_offset(self):
        size = self._size * self._scale
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        # Scaling
        width = self._size[0] * self._scale[0]
        height = self._size[1] * self._scale[1]
        size = (int(width), int(height))
        color = _process_color(self._color)
        if size[0] <= 0 or size[1] <= 0:
            target = InternalImage(size=(1, 1)).fill(color)
            self._transform_image = target._surf
            self._recalculate_offset()
            self._expire_static()
            return
        new_image = InternalImage(size=size)
        new_image.draw_arc(color,
                           math.pi/180 * self._start_angle,
                           math.pi/180 * self._stop_angle,
                           (0, 0), self._size, self._border)
        # Flip
        if self._flip_x or self._flip_y:
            new_image.flip(self._flip_x, self._flip_y)
        # Rotation
        if self._angle != 0:
            angle = self._angle % 360
            old = Vec2D(new_image.rect.center)
            new_image.rotate(angle)
            new = new_image.rect.center
            self._transform_offset = old - new
        self._transform_image = new_image._surf
        self._recalculate_offset()
        self._expire_static()

    @property
    def start_angle(self):
        return self._start_angle

    @start_angle.setter
    def start_angle(self, value):
        self._start_angle = value
        self._redraw_internal_image()

    @property
    def stop_angle(self):
        return self._stop_angle

    @stop_angle.setter
    def stop_angle(self, value):
        self._stop_angle = value
        self._redraw_internal_image()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._redraw_internal_image()

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        self._redraw_internal_image()


def arc(color, start_angle, stop_angle, width, height,
        x=None, y=None, thickness=1, anchor='center', **kwargs):
    """
    Function to make an arc.

    :param color: color to draw arc
    :type color: str or List[str]
    :param start_angle: angle to start drawing arc at
    :type start_angle: int
    :param stop_angle: angle to stop drawing arc at
    :type stop_angle: int
    :param thickness: thickness of arc in pixels
    :type thickness: int
    :param args: left top corner of arc and width and height of arc
    :type args: two Tuples (left, top), (width, height) or four ints left, top, width, height
    :return: Arc object created
    """
    return Arc((x, y), width, height, start_angle, stop_angle, anchor, color, thickness, **kwargs)


from os import environ

environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

import pygame
from designer.core.director import *
from designer.core.event import *
from designer.helpers import *
from designer.animation import *
from designer.utilities.easings import *
from designer.objects import *
from designer.colors import *
from designer.positioning import *
from designer.keyboard import *
from designer.mouse import *
from designer.movement import *

GLOBAL_DIRECTOR: Director = None

__all__ = [
    'circle', 'ellipse',
    'arc', 'line',
    'rectangle',
    'text',
    'shape', 'lines', 'pen',
    'background_image',
    'image', 'emoji',
    'group',
    'draw',
    # Window information
    'set_window_color', 'get_window_color',
    'set_window_size',
    'get_height', 'get_window_height',
    'get_width', 'get_window_width',
    # Events
    'when', 'starting', 'updating', 'typing', 'clicking',
    'start',
    'stop',
    'pause',
    'colliding', 'colliding_with_mouse',
    'destroy',
    'DesignerObject',
    # Positioning
    'above', 'below',
    # Director stuff
    'get_director',
    # Window stuff
    'set_window_title', 'get_window_title', 'set_window_image',
    # Keyboard stuff
    'get_keyboard_repeat', 'set_keyboard_repeat',
    'get_keyboard_delay', 'set_keyboard_delay',
    'get_keyboard_interval', 'set_keyboard_interval',
    'enable_keyboard_repeating', 'disable_keyboard_repeating',
    # Mouse stuff
    'get_mouse_cursor', 'set_mouse_cursor',
    'get_mouse_visible', 'set_mouse_visible',
    'get_mouse_position', 'set_mouse_position',
    'get_mouse_x', 'get_mouse_y',
    # Animations
    'Animation', 'linear_animation', 'sequence_animation',
    'glide_around',
    'glide_right',
    'glide_left',
    'glide_up',
    'glide_down',
    'glide_in_degrees',
    'spin',
    # Easings
    'Linear', 'Iterate',
    # Music
    'play_sound',
    'play_music', 'background_music', 'pause_music', 'set_music_volume', 'is_music_playing',
    'get_music_volume', 'stop_music', 'rewind_music', 'continue_music', 'set_music_position', 'get_music_position',
    # Movement
    'move_forward', 'move_backward', 'turn_left', 'turn_right', 'go_to', 'go_to_xy', 'go_to_mouse',
    'point_towards', 'point_towards_mouse', 'point_in_direction', 'change_xy', 'change_x', 'change_y', 'set_x', 'set_y',
    'get_angle', 'get_x', 'get_y',
    'flip_x', 'flip_y', 'set_flip_x', 'set_flip_y', 'set_scale', 'set_scale_x', 'set_scale_y', 'set_background_image',
    'get_scale', 'get_scale_x', 'get_scale_y', 'get_visible', 'get_flip_x', 'get_flip_y', 'show', 'hide',
    'grow', 'grow_x', 'grow_y', 'shrink',
    'move_to_x', 'move_to_y', 'move_to', 'move_to_mouse', 'move_to_xy',
    'set_visible', 'change_scale',
    # Emoji specific
    'get_emoji_name', 'set_emoji_name'
]



from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities import Vec2D
from designer.utilities.util import _anchor_offset


class Circle(DesignerObject):
    DEFAULT_BORDER_WIDTH = 1
    FIELDS = (*DesignerObject.FIELDS, 'radius', 'color', 'border')

    def __init__(self, center, anchor, radius, color, border):
        """
        Creates a circle object on the window.

        :param center: x, y coordinates of center of circle
        :type center: Tuple[int]
        :param anchor: the anchor to draw the circle at
        :type anchor: str
        :param radius: The radius of the circle
        :type radius: int
        :param color: color of circle
        :type color: str or List[str]
        :param border: the width of the circle's line (0 is used for a filled circle)
        :type border: int
        """
        super().__init__()

        x, y = center
        x = x if x is not None else get_width()/2
        y = y if y is not None else get_height()/2
        center = x, y

        self._pos = center
        self._anchor = anchor
        # Circle specific data
        self._radius = radius
        self._color = color
        self._border = border

        # And draw!
        self._redraw_internal_image()

    def __repr__(self):
        activated = "" if self._active else "INACTIVE "
        return f"<{activated}circle({self._color!r}, {self._radius})>"

    def _recalculate_offset(self):
        size = 2 * self._radius * self.scale[0]
        offset = _anchor_offset(self._anchor, size, size)
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        radius = self._radius * self._scale[0]
        diameter = 2 * radius
        color = _process_color(self._color)
        if int(radius) > 0:
            new_image = InternalImage(size=(diameter, diameter))
            new_image.draw_circle(color, (radius, radius),
                                  radius, self._border or 0)
            target = new_image
        else:
            target = InternalImage(size=(1, 1))
            target.fill(color)
        self._transform_image = target._surf
        self._recalculate_offset()
        self._expire_static()

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = value
        self._redraw_internal_image()

    @property
    def size(self):
        return Vec2D(self._radius, self._radius)

    @size.setter
    def size(self, value):
        if isinstance(value, (int, float)):
            self._radius = int(value)
        else:
            self._radius = value[0]
        self._redraw_internal_image()

    @property
    def width(self):
        return self._radius*2

    @width.setter
    def width(self, value):
        self._radius = value*2
        self._redraw_internal_image()

    @property
    def height(self):
        return self._radius*2

    @height.setter
    def height(self, value):
        self._radius = value*2
        self._redraw_internal_image()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._redraw_internal_image()

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        self._redraw_internal_image()


def circle(color, radius, x=None, y=None, anchor='center', border=None):
    """
    Function to create a circle.

    >>> centered_circle = circle('red', 10)
    >>> positioned_circle = circle('blue', 20, 0, 0, 'topleft')
    >>> empty_circle = circle('black', 10, filled=False)

    :param x:
    :param y:
    :param filled:
    :param border:
    :param anchor:
    :param color: color of circle
    :param radius: int, radius of circle in pixels
    :return: Circle object created
    """
    if not isinstance(radius, (int, float)):
        raise ValueError(f"The parameter radius was given the value {radius!r} but that parameter must be either an integer or a float.")
    if x is not None and y is None:
        x, y = x
    return Circle((x, y), anchor, radius, color, border)



import pygame

from designer.objects.designer_object import DesignerObject
from designer.objects.group import group


def above(top: DesignerObject, bottom: DesignerObject):
    """ Moves the bottom to be below the top """
    bottom.x = top.x + top.width/2 - bottom.width/2
    bottom.y = top.y + top.height
    return group(top, bottom)

def below(bottom: DesignerObject, top: DesignerObject, ):
    """ Moves the bottom to be below the top """
    bottom.x = top.x + abs(top.width - bottom.width)//2
    bottom.y = top.y + top.height
    return group(top, bottom)



import pygame

from collections import deque

from designer import register
from designer.objects.designer_object import DesignerObject
from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities.vector import Vec2D
from designer.utilities.util import _anchor_offset, rect_from_points
from designer.utilities.animation import *
from designer.utilities.easings import Linear, Iterate, Random, LinearTuple
from designer.utilities.rect import Rect


class Pen(DesignerObject):
    """
    TODO: Finish this with the idea of iteratively "painting" on a canvas.

    """

    def __init__(self, center, thickness, color, speed):
        super().__init__()

        x, y = center
        x = x if x is not None else get_width() / 2
        y = y if y is not None else get_height() / 2
        center = Vec2D(x, y)

        self._pos = center
        # Pen specific data
        self._color = color
        self._thickness = thickness
        self._inking = True
        self._speed = speed
        # The points we have visited
        self._pen_pos = center
        self._visited_points = []
        self._bounds = self._get_bounds()
        self._size = self._bounds.size
        self._current_animation = None
        self._queued_animations = deque()
        self._speed = 4

        # And draw!
        self._redraw_internal_image()

        register('Pen.pen_pos.animation.start', self._visit_point)
        register('Pen.pen_pos.animation.end', self._finish_point)

    def _restart_animation_if_needed(self):
        if not self._current_animation and self._queued_animations:
            x, y, speed = self._queued_animations.popleft()
            print(self._pen_pos, (x, y), speed)
            self._current_animation = Animation('pen_pos', LinearTuple(self._pen_pos, (x, y)), self._speed)
            self._speed += 1
            self.animate(self._current_animation)

    @property
    def pen_pos(self):
        """
        The position of a sprite in 2D coordinates, represented as a
        :class:`Vec2D <spyral.Vec2D>`
        """
        return self._pen_pos

    @pen_pos.setter
    def pen_pos(self, value):
        if value == self._pen_pos:
            return
        self._pen_pos = Vec2D(value)
        self._bounds = self._get_bounds()
        self._pos = self._bounds.topleft
        self._size = self._bounds.size
        self._redraw_internal_image()
        self._expire_static()

    def move_to(self, x, y, speed=None):
        if speed is None:
            speed = self._speed
        self._queued_animations.append((x, y, speed))
        self._restart_animation_if_needed()

    def _visit_point(self):
        self._visited_points.append(Vec2D(self._pen_pos))
        self._bounds = self._get_bounds()
        self._size = self._bounds.size
        self._redraw_internal_image()

    def _finish_point(self):
        self._current_animation = None
        self._restart_animation_if_needed()

    def _get_bounds(self):
        if not self._visited_points:
            return Rect(self._pen_pos.x, self._pen_pos.y, 1, 1)
        bounds = rect_from_points(self._visited_points + [self._pen_pos])
        # TODO: Refine this calculation of border expansion
        if self._thickness > 0:
            bounds.inflate_ip(self._thickness*2, self._thickness*2)
        return bounds

    def _recalculate_offset(self):
        if self._transform_image is None:
            return
        size = self._scale * self._transform_image.get_size()
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        # TODO: This should use math to calculate the angle/flip/scale instead of the pygame objects
        target = InternalImage(size=self._size)
        points = self._visited_points + [self._pen_pos]
        if len(points) <= 1:
            return self._make_blank_surface()
        origin = self._bounds.topleft
        # TODO: Refine this calculation of border expansion
        if self._thickness > 1:
            origin = origin + Vec2D(self._thickness, self._thickness)
        offset_points = [point-origin for point in points]
        pygame.draw.lines(target._surf, _process_color(self._color), False,
                            offset_points, self._thickness)
        self._default_redraw_transforms(target)


def pen(color='black', x=None, y=None, thickness=1, speed=2):
    return Pen((x, y), thickness, color, speed)



import os
import io
from urllib.request import urlopen, Request
import pygame


class MusicModule:
    _USER_AGENT = "Designer Game Library for Python"

    def __init__(self):
        self._filename = None
        self._volume = 1.0
        # pygame.mixer.music.set_endevent
        # Use this to get events about music!

    def _load_song(self):
        try:
            path_strs = self._filename.split('/')
            fixed_paths = os.path.join(*path_strs)
            if os.path.exists(fixed_paths):
                pygame.mixer.music.load(self._filename)
            else:
                raise FileNotFoundError(fixed_paths)
        except FileNotFoundError as err:
            try:
                req = Request(self._filename, headers={'User-Agent': self._USER_AGENT})
                with urlopen(req) as opened_song:
                    song_str = opened_song.read()
                    song_file = io.BytesIO(song_str)
                    pygame.mixer.music.load(song_file)
            except:
                if self._filename.startswith('https://') or self._filename.startswith('http://'):
                    raise ValueError(f"Unexpected error while playing url: {self._filename!r}")
                raise ValueError(f"Unexpected error while playing song from filename: {self._filename!r}")

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value
        self._load_song()
        self.play()

    @property
    def playing(self):
        return pygame.mixer.music.get_busy()

    @property
    def volume(self):
        return pygame.mixer.music.get_volume()

    @volume.setter
    def volume(self, value):
        pygame.mixer.music.set_volume(value)

    def play(self, loop=True):
        pygame.mixer.music.play(loops=loop)

    def stop(self):
        pygame.mixer.music.stop()

    def pause(self):
        pygame.mixer.music.pause()

    def unpause(self):
        pygame.mixer.music.unpause()

    def rewind(self):
        pygame.mixer.music.rewind()

    def set_time_position(self, time):
        pygame.mixer.music.set_pos(time)

    def get_time_position(self):
        return pygame.mixer.music.get_pos()


"""Clock to manage frame/cpu rate."""

# spyral uses a modified version of this file from Gummworld2
# and Trolls Outta Luckland, and thanks the authors for their great,
# reusable piece of code. It is licensed separately from spyral
# under the LGPLv3 or newer.

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.


# CREDITS
#
# The inspiration for this module came from Koen Witters's superb article
# "deWiTTERS Game Loop", aka "Constant Game Speed independent of Variable FPS"
# at http://www.koonsolo.com/news/dewitters-gameloop/.
#
# The clock was changed to use a fixed time-step after many discussions with
# DR0ID, and a few readings of
# http://gafferongames.com/game-physics/fix-your-timestep/.
#
# Thanks to Koen Witters, DR0ID, and Glenn Fiedler for sharing.
#
# Pythonated by Gummbum. While the builtin demo requires pygame, the module
# does not. The GameClock class is purely Python and should be compatible with
# other Python-based multi-media and game development libraries.


__version__ = '$Id: gameclock.py$'
__author__ = 'Gummbum, (c) 2011-2012'

__doc__ = """

GameClock is a fixed time-step clock that keeps time in terms of game
time. It will attempt to keep game time close to real time, so if an
interval takes one second of game time then the user experiences one
second of real time. In the worst case where the CPU cannot keep up with
the game load, game time will subjectively take longer but still remain
accurate enough to keep game elements in synchronization.

GameClock manages time in the following ways:

    1.  Register special callback functions that will be run when they
        are due.
    2.  Schedule game logic updates at a constant speed, independent of
        frame rate.
    3.  Schedule frames at capped frames-per-second, or uncapped.
    4.  Invoke a pre-emptive pause callback when paused.
    5.  Schedule miscellaneous items at user-configurable intervals.
    6.  Optionally sleep between schedules to conserve CPU cycles.
    7.  Gracefully handles corner cases.

Note the Python Library docs mention that not all computer platforms'
time functions return time fractions of a second. This module will not
work on such platforms.

USAGE

Callback:

    clock = GameClock(
        update_callback=update_world,
        frame_callback=draw_scene,
        pause_callback=pause_game)
    while 1:
        clock.tick()

Special callbacks can be directly set and cleared at any time:

    clock.update_callback = my_new_update
    clock.frame_callback = my_new_draw
    clock.pause_callback = my_new_pause

    clock.update_callback = None
    clock.frame_callback = None
    clock.pause_callback = None

Scheduling miscellanous callbacks:

    def every_second_of_every_day(dt):
        "..."
    clock.schedule_interval(every_second_of_every_day, 1.0)

The update_callback receives a single DT argument, which is the time-step
in seconds since the last update.

The frame_callback receives a single INTERPOLATION argument, which is the
fractional position in time of the frame within the current update time-
step. It is a float in the range 0.0 to 1.0.

The pause_callback receives no arguments.

User-defined interval callbacks accept at least a DT argument, which is
the scheduled item's interval, and optional user-defined arguments. See
GameClock.schedule_interval.

"""

import time


class _IntervalItem(object):
    """An interval item runs after an elapsed interval."""
    __slots__ = ['func', 'interval', 'lasttime', 'life', 'args']

    def __init__(self, func, interval, curtime, life, args):
        self.func = func
        self.interval = float(interval)
        self.lasttime = curtime
        self.life = life
        self.args = args


class GameClock:
    """
    GameClock is an implementation of fixed-timestep clocks used for
    running the game.

    =============== ============
    Attribute       Description
    =============== ============
    get_ticks       The time source for the game. Should support at least
                    subsecond accuracy.
    max_ups         The maximum number of updates per second. The clock
                    will prioritize trying to keep the number of
                    updates per second at this level, at the cost of
                    frames per second
    max_fps         The maximum number of frames per second.
    use_wait        Boolean which represents whether the clock should
                    try to sleep when possible. Setting this to False
                    will give better game performance on lower end
                    hardware, but take more CPU power
    update_callback The function which should be called when for
                    updates
    frame_callback  The function which should be called for frame
                    rendering
    game_time       Virtual elapsed time in milliseconds
    paused          The game time at which the clock was paused
    =============== ============

    In addition to these attributes, the following read-only attributes
    provide some useful metrics on performance of the program.

    =============== ============
    Attribute       Description
    =============== ============
    num_updates     The number of updates run in the current one-second
                    interval
    num_frames      The number of frames run in the current one-second
                    interval
    dt_update       Duration of the previous update
    dt_frame        Duration of the previous update
    cost_of_update  How much real time the update callback takes
    cost_of_frame   How much real time the frame callback takes
    ups             Average number of updates per second over the last
                    five seconds
    fps             Average number of frames per second over the last
                    five seconds
    =============== ============
    """

    def __init__(self,
                 max_ups=30,
                 max_fps=0,
                 use_wait=False,
                 time_source=time.time,
                 update_callback=None,
                 frame_callback=None,
                 paused_callback=None):

        # Configurables.
        self.get_ticks = time_source
        self.max_ups = max_ups
        self.max_fps = max_fps
        self.use_wait = use_wait
        self.update_callback = update_callback
        self.frame_callback = frame_callback
        self.paused_callback = paused_callback

        # Time keeping.
        CURRENT_TIME = self.get_ticks()
        self._real_time = CURRENT_TIME
        self._game_time = CURRENT_TIME
        self._last_update = CURRENT_TIME
        self._last_update_real = CURRENT_TIME
        self._next_update = CURRENT_TIME
        self._last_frame = CURRENT_TIME
        self._next_frame = CURRENT_TIME
        self._next_second = CURRENT_TIME
        self._update_ready = False
        self._frame_ready = False
        #        self._frame_skip = 0
        self._paused = 0

        # Schedules
        self._need_sort = False
        self._schedules = []
        self._unschedules = []

        # Metrics: update and frame progress counter in the current one-second
        # interval.
        self.num_updates = 0
        self.num_frames = 0
        # Metrics: duration in seconds of the previous update and frame.
        self.dt_update = 0.0
        self.dt_frame = 0.0
        # Metrics: how much real time a callback consumes
        self.cost_of_update = 0.0
        self.cost_of_frame = 0.0
        # Metrics: average updates and frames per second over the last five
        # seconds.
        self.ups = 0.0
        self.fps = 0.0

    @property
    def max_ups(self):
        return self._max_ups

    @max_ups.setter
    def max_ups(self, val):
        self._max_ups = val
        self._update_interval = 1.0 / val

    @property
    def max_fps(self):
        return self._max_fps

    @max_fps.setter
    def max_fps(self, val):
        self._max_fps = val
        self._frame_interval = 1.0 / val if val > 0 else 0

    @property
    def game_time(self):
        return self._game_time

    @property
    def paused(self):
        return self._paused

    @property
    def interpolate(self):
        interp = (self._real_time - self._last_update_real)
        interp = interp / self._update_interval
        return interp if interp <= 1.0 else 1.0

    def tick(self):
        """
        Should be called in a loop, will run and handle timers.
        """
        # Now.
        real_time = self.get_ticks()
        self._real_time = real_time

        # Pre-emptive pause callback.
        if self._paused:
            if self.paused_callback:
                self.paused_callback()
            return

        # Check if update and frame are due.
        update_interval = self._update_interval
        game_time = self._game_time
        if real_time >= self._next_update:
            self.dt_update = real_time - self._last_update_real
            self._last_update_real = real_time
            game_time += update_interval
            self._game_time = game_time
            self._last_update = game_time
            self._next_update = real_time + update_interval
            self.num_updates += 1
            if self.update_callback:
                self._update_ready = True
        # ORIG
        #        if (real_time + self.cost_of_frame < self._next_update) and
        #            (real_time >= self._next_frame):
        #            self.dt_frame = real_time - self._last_frame
        #            self._last_frame = real_time
        #            self._next_frame = real_time + self._frame_interval
        #            self.num_frames += 1
        #            if self.frame_callback:
        #                self._frame_ready = True
        # END ORIG
        #
        # SACRIFICE FRAMES TO MAINTAIN UPDATES
        if real_time >= self._next_frame:
            do_frame = False
            if real_time + self.cost_of_frame <= self._next_update:
                do_frame = True
            elif self._frame_skip > 0:
                do_frame = True
            else:
                self._next_frame = self._next_update
                self._frame_skip += 1
            if do_frame:
                self._frame_skip = 0
                self.dt_frame = real_time - self._last_frame
                self._last_frame = real_time
                self._next_frame = real_time + self._frame_interval
                self.num_frames += 1
                if self.frame_callback:
                    self._frame_ready = True
        # END SACRIFICE FRAMES TO MAINTAIN UPDATES
        #
        if real_time - self._last_frame >= self._update_interval or (
                real_time + self.cost_of_frame < self._next_update and
                real_time >= self._next_frame):
            self.dt_frame = real_time - self._last_frame
            self._last_frame = real_time
            self._next_frame = real_time + self._frame_interval
            self.num_frames += 1
            if self.frame_callback:
                self._frame_ready = True

        # Check if a schedule is due, and when.
        sched_ready = False
        sched_due = 0
        if self._schedules:
            sched = self._schedules[0]
            sched_due = sched.lasttime + sched.interval
            if real_time >= sched_due:
                sched_ready = True

        # Run schedules if any are due.
        if self._update_ready or sched_ready:
            self._run_schedules()

        # Run the frame callback (moved inline to reduce function calls).
        if self.frame_callback and self._frame_ready:
            get_ticks = self.get_ticks
            ticks = get_ticks()
            self.frame_callback(self.interpolate)
            self.cost_of_frame = get_ticks() - ticks
            self._frame_ready = False

        # Flip metrics counters every second.
        if real_time >= self._next_second:
            self._flip(real_time)

        # Sleep to save CPU.
        if self.use_wait:
            upcoming_events = [
                self._next_frame,
                self._next_update,
                self._next_second,
            ]
            if sched_due != 0:
                upcoming_events.append(sched_due)
            next = min(upcoming_events)
            ticks = self.get_ticks()
            time_to_sleep = next - ticks
            if time_to_sleep >= 0.002:
                time.sleep(time_to_sleep)

    def pause(self):
        """Pause the clock so that time does not elapse.

        While the clock is paused, no schedules will fire and tick() returns
        immediately without progressing internal counters. Game loops that
        completely rely on the clock will need to take over timekeeping and
        handling events; otherwise, the game will appear to deadlock. There are
        many ways to solve this scenario. For instance, another clock can be
        created and used temporarily, and the original swapped back in and
        resumed when needed.
        """
        self._paused = self.get_ticks()

    def resume(self):
        """Resume the clock from the point that it was paused."""
        real_time = self.get_ticks()
        paused = self._paused
        for item in self._schedules:
            delta = paused - item.lasttime
            item.lasttime = real_time - delta
        self._last_update_real = real_time - (paused - self._last_update_real)
        self._paused = 0
        self._real_time = real_time

    def schedule_interval(self, func, interval, life=0, args=None):
        """

        Schedule an item to be called back each time an interval elapses.

        While the clock is paused time does not pass.

        | *func*: The callback function.
        | *interval*: The time in seconds (float) between calls.
        | *life*: The number of times the callback will fire, after which the
          schedule will be removed. If the value 0 is specified, the event
          will persist until manually unscheduled.
        | *args*: A list that will be passed to the callback as an unpacked
          sequence, like so: item.func(\*[item.interval]+item.args).

        """
        if args is None:
            args = []
        self.unschedule(func)
        item = _IntervalItem(
            func, interval, self.get_ticks(), life, [interval] + list(args))
        self._schedules.append(item)
        self._need_sort = True

    def unschedule(self, func):
        """Unschedule a managed function."""
        sched = self._schedules
        for item in list(sched):
            if item.func == func:
                sched.remove(item)

    @staticmethod
    def _interval_item_sort_key(item):
        return item.lasttime + item.interval

    def _run_schedules(self):
        get_ticks = self.get_ticks

        # Run the update callback.
        if self.update_callback and self._update_ready:
            t = get_ticks()
            self.update_callback(self.dt_update)
            self.cost_of_update = get_ticks() - t
            self._update_ready = False

        # Run the interval callbacks.
        if self._need_sort:
            self._schedules.sort(key=self._interval_item_sort_key)
            self._need_sort = False
        real_time = self._real_time
        for sched in self._schedules:
            interval = sched.interval
            due = sched.lasttime + interval
            if real_time >= due:
                sched.func(*sched.args)
                sched.lasttime += interval
                need_sort = True
                if sched.life > 0:
                    if sched.life == 1:
                        self._unschedules.append(sched.func)
                        need_sort = False
                    else:
                        sched.life -= 1
                if need_sort:
                    self._need_sort = True
            else:
                break
        if self._unschedules:
            for func in self._unschedules:
                self.unschedule(func)
            del self._unschedules[:]

    def _flip(self, real_time):
        self.ups = self.num_updates
        self.fps = self.num_frames

        self.num_updates = 0
        self.num_frames = 0

        self._last_second = real_time
        self._next_second += 1.0




from designer.utilities.rect import Rect
from designer.utilities.vector import Vec2D
from designer.utilities.util import *
from designer.utilities.file_system import *
from designer.utilities.memoize import *



import designer
import os


def get_designer_path():
    """ Get the path of the Designer directory. """
    return os.path.dirname(designer.__file__) + '/'


def get_resource(filename: str):
    """
    Get the given resource by its filename
    """
    return os.path.join(os.path.dirname(designer.__file__), 'data', filename)
